# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Callable, List, Optional, Tuple

import torch

from atek.data_preprocess.atek_data_sample import MultiFrameCameraData
from atek.util.camera_calib_utils import (
    rescale_pixel_coords,
    rotate_pixel_coords_cw90,
    undistort_pixel_coords,
)

from omegaconf.omegaconf import DictConfig
from PIL import Image
from projectaria_tools.core import calibration, data_provider
from projectaria_tools.core.sensor_data import TimeDomain, TimeQueryOptions  # @manual
from torchvision.transforms import InterpolationMode, v2

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def identity_transform(obj):
    """Helper identity transform function"""
    return obj


class AriaCameraProcessor:
    def __init__(
        self,
        video_vrs: str,
        conf: DictConfig,  # TODO: consider use more explicit init, and overload with DictConfig
    ):
        # Parse in conf
        self.conf = conf
        # Resolution-rescale related params
        self.target_camera_resolution: List = (
            conf.target_camera_resolution if "target_camera_resolution" in conf else []
        )
        self.rescale_antialias: bool = (
            conf.rescale_antialias if "rescale_antialias" in conf else True
        )
        # Camera model undistortion related params
        self.undistort_to_linear_camera: bool = (
            conf.undistort_to_linear_camera
            if "undistort_to_linear_camera" in conf
            else False
        )
        # Image rotation related params
        self.rotate_image_cw90deg: bool = (
            conf.rotate_image_cw90deg if "rotate_image_cw90deg" in conf else False
        )

        # setting up vrs data provider
        self.video_vrs = video_vrs
        self.camera_label = conf.sensor_label
        self.data_provider, self.stream_id = self.setup_vrs_data_provider()

        # setting up camera calibration, and optional linear camera (for rectification)
        self.camera_calibration = self.setup_camera_calibration()

        # Cache capture timestamps
        self.time_domain = getattr(TimeDomain, conf.time_domain)
        self.camera_timestamps = self.data_provider.get_timestamps_ns(
            self.stream_id, self.time_domain
        )

    def get_final_camera_calib(self):
        return self.final_camera_calib

    def get_stream_id(self):
        return self.stream_id

    def get_origin_label(self):
        """
        returns the sensor label of the origin (DeviceFrame) definition in Aria calibration
        """
        return self.data_provider.get_device_calibration().get_origin_label()

    def setup_vrs_data_provider(self):
        """
        Setup the vrs data provider, and only activate the stream id specified by camera_label.
        Returns: vrs_data_provider, stream_id
        """
        provider = data_provider.create_vrs_data_provider(self.video_vrs)
        assert (
            provider is not None
        ), f"Cannot open video.vrs file under path [{self.video_vrs}]"
        stream_id = provider.get_stream_id_from_label(self.camera_label)
        assert (
            stream_id is not None
        ), f"Cannot find stream id for camera [{self.camera_label}]"

        return provider, stream_id

    def setup_camera_calibration(
        self,
    ):
        """
        Setup camera calibration from the VRS, which is done through the following steps (IN ORDER!):
        1. obtain camera calibration from vrs, this corresponds to the VRS resolution.
        2. (if `undistort_to_linear_camera` is True) undistort the camera calibration to a linear camera calibration,
                where focal is kept the same as `focal_x` value in the original intrinsics.
        3. (if `target_camera_resolution` or `target_linear_camera_focal` are specified) convert camera calibration to a linear camera calibration
        4. (if `rotate_image_cw90deg` is True) rotate the camera calibration (both intrinsics and extrinsics) by 90deg clockwise.
            Note that the `target_camera_resolution` and `target_linear_camera_params` should be set according to the UNROTATED camera.
        """
        # obtain the camera calibration from vrs
        stream_sensor_calib = self.data_provider.get_sensor_calibration(self.stream_id)
        assert (
            stream_sensor_calib is not None
        ), f"Can not get the sensor calibration for {self.stream_id} in vrs {self.video_vrs}"
        camera_calib = stream_sensor_calib.camera_calibration()
        self.original_camera_calib = camera_calib

        # undistort to linear camera if specified
        if self.undistort_to_linear_camera:
            camera_calib = calibration.get_linear_camera_calibration(
                image_width=camera_calib.get_image_size()[0],
                image_height=camera_calib.get_image_size()[1],
                focal_length=camera_calib.get_focal_lengths()[0],
                label=f"undistorted_linear_{camera_calib.get_label()}",
                T_Device_Camera=camera_calib.get_transform_device_camera(),
            )
            self.undistorted_linear_camera_calib = (
                camera_calib  # save this for image undistortion operations
            )

        # rescale resolution if specified
        if (
            self.target_camera_resolution is not None
            and len(self.target_camera_resolution) == 2
        ):
            self.scale = (
                float(self.target_camera_resolution[0])
                / camera_calib.get_image_size()[0]
            )
            camera_calib = camera_calib.rescale(
                new_resolution=self.target_camera_resolution, scale=self.scale
            )

        # Rotate camera model if specified
        if self.rotate_image_cw90deg:
            camera_calib = calibration.rotate_camera_calib_cw90deg(camera_calib)

        self.final_camera_calib = camera_calib

    class DistortByCalibrationTVWrapper:
        """
        A torch vision transform function wrapper on top of `calibration.distort_by_calibration()` function,
        so that this operation can be chained with other tv.transforms.
        """

        def __init__(self, dstCalib, srcCalib, is_transforming_label_data=False):
            self.dstCalib = dstCalib
            self.srcCalib = srcCalib
            self.is_transforming_label_data = is_transforming_label_data

        def __call__(self, image):
            if image.ndim != 4:
                raise ValueError(
                    f"Expecting 4D tensor of [Frame, C, H, W], got {image.ndim}D tensor instead."
                )
            if image.dtype in [torch.int32]:
                # projectaria_tools calibration will cast int32 or uint32 to uint8 but leaves uint64 as uint64
                image = image.to(torch.uint64)

            num_frames = image.size(0)

            result_list = []
            for i in range(num_frames):
                # input image is tensor shape of [Frame, C, H, W], while distort_by_calibration requires [H, W, C]
                single_image = image[i].permute(1, 2, 0).numpy().copy()

                if self.is_transforming_label_data:
                    numpy_result = calibration.distort_label_by_calibration(
                        single_image, self.dstCalib, self.srcCalib
                    )
                else:
                    numpy_result = calibration.distort_by_calibration(
                        single_image, self.dstCalib, self.srcCalib
                    )
                tensor_result = torch.from_numpy(numpy_result)
                if tensor_result.ndim == 2:
                    # [H, W] -> [1, H, W]
                    tensor_result = tensor_result.unsqueeze(0)
                else:
                    # [H, W, C] -> [C, H, W]
                    tensor_result = tensor_result.permute(2, 0, 1)

                result_list.append(tensor_result)

            return torch.stack(result_list, dim=0)

    def get_image_transform(
        self,
        rescale_interpolation: InterpolationMode = InterpolationMode.BILINEAR,
        is_transforming_label_data: bool = False,
    ) -> Callable:
        """
        Returns a list of torchvision transform functions to be applied to the raw image data,
        in the order of: undistortion -> rescale -> rotateCW90, where any step is optional.
        If transforming label data (as is the case with segmentation masks), the interpolation
        should be set to nearest, and is_transforming_label_data should be set to True.
        """
        image_transform_list = []
        # undistort if specified
        if self.undistort_to_linear_camera:
            image_transform_list.append(
                self.DistortByCalibrationTVWrapper(
                    dstCalib=self.undistorted_linear_camera_calib,
                    srcCalib=self.original_camera_calib,
                    is_transforming_label_data=is_transforming_label_data,
                )
            )

        if (
            self.target_camera_resolution is not None
            and len(self.target_camera_resolution) == 2
        ):
            # resolution is specified as [w, h]. Need to pass [h,w] here
            image_transform_list.append(
                v2.Resize(
                    [
                        self.target_camera_resolution[1],
                        self.target_camera_resolution[0],
                    ],
                    interpolation=rescale_interpolation,
                    antialias=self.rescale_antialias,
                )
            )

        if self.rotate_image_cw90deg:
            # image is [frames, c, h, w]
            image_transform_list.append(lambda img: torch.rot90(img, k=3, dims=[2, 3]))

        if len(image_transform_list) == 0:
            return identity_transform
        else:
            return v2.Compose(image_transform_list)

    def get_pixel_transform(self) -> Callable:
        """
        Returns a list of torchvision transform functions to be applied to the raw image data,
        in the order of: undistortion -> rescale -> rotateCW90, where any step is optional.
        The transform should be applied to pixel coordinates of Tensor [N, 2]
        """
        pixel_transform_list = []
        # undistort if specified
        if self.undistort_to_linear_camera:
            pixel_transform_list.append(
                lambda pixels: undistort_pixel_coords(
                    pixels,
                    src_calib=self.original_camera_calib,
                    dst_calib=self.undistorted_linear_camera_calib,
                )
            )

        # rescale resolution if specified
        if (
            self.target_camera_resolution is not None
            and len(self.target_camera_resolution) == 2
        ):

            pixel_transform_list.append(
                lambda pixels: rescale_pixel_coords(pixels, scale=self.scale)
            )

        # rotate if specified
        if self.rotate_image_cw90deg:
            # image is [frames, c, h, w]
            pixel_transform_list.append(
                lambda pixels: rotate_pixel_coords_cw90(
                    pixels, image_dim_after_rot=self.final_camera_calib.get_image_size()
                )
            )

        if len(pixel_transform_list) == 0:
            return identity_transform
        else:
            return v2.Compose(pixel_transform_list)

    def get_image_data_by_timestamps_ns(
        self, timestamps_ns: List[int]
    ) -> Optional[MultiFrameCameraData]:
        """
        Obtain images by timestamps. Image should be processed.
        returns: if successful, returns (image_data: Tensor [numFrames, numChannel, height, width], capture_timestamp: Tensor[numFrames], frame_id_in_stream: Tensor[numFrames])
                else returns None
        """
        image_list = []
        capture_timestamp_list = []
        frame_id_list = []
        exposure_list = []
        gain_list = []
        for single_timestamp in timestamps_ns:
            index = self.data_provider.get_index_by_time_ns(
                stream_id=self.stream_id,
                time_ns=single_timestamp,
                time_domain=self.time_domain,
                time_query_options=TimeQueryOptions.CLOSEST,
            )
            image_data_and_record = self.data_provider.get_image_data_by_index(
                self.stream_id, index
            )
            frame_id = image_data_and_record[1].frame_number
            capture_timestamp = image_data_and_record[1].capture_timestamp_ns

            # Check if fetched frame is within tolerance
            if abs(capture_timestamp - single_timestamp) > self.conf.tolerance_ns:
                continue

            # reshape image to proper tensor shape
            image = torch.from_numpy(image_data_and_record[0].to_numpy_array())

            if len(image.shape) == 2:
                # single channel image: [h,w] -> [c, h, w]
                image = torch.unsqueeze(image, dim=0)
            else:
                # rgb image: [h, w, c] -> [c, h ,w]
                image = image.permute(2, 0, 1)
            image_list.append(image)

            # insert other values from the image data
            image_record = image_data_and_record[1]
            exposure_list.append(image_record.exposure_duration)
            gain_list.append(image_record.gain)
            capture_timestamp_list.append(capture_timestamp)
            frame_id_list.append(frame_id)
        # End for single_timestamp

        # Check if at least one frame is successfully fetched
        if len(image_list) == 0:
            return None

        # Image transformations are handled by torchvision's transform functions.
        batched_image_tensor = torch.stack(image_list, dim=0)

        image_transform = self.get_image_transform()
        batched_image_tensor = image_transform(batched_image_tensor)

        # properly clean output to desired dtype and shapes
        result = MultiFrameCameraData(
            images=batched_image_tensor,
            capture_timestamps_ns=torch.tensor(
                capture_timestamp_list, dtype=torch.int64
            ),
            frame_ids=torch.tensor(frame_id_list, dtype=torch.int64),
            exposure_durations_s=torch.tensor(exposure_list, dtype=torch.float32),
            gains=torch.tensor(gain_list, dtype=torch.float32),
        )

        return result



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import random
from typing import Dict, Optional

import numpy as np
import torch

import trimesh

from atek.evaluation.surface_reconstruction.surface_reconstruction_utils import (
    compute_pts_to_mesh_dist,
    correct_adt_mesh_gravity,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def evaluate_single_mesh_pair(
    pred_mesh_filename: str,
    gt_mesh_filename: str,
    correct_mesh_gravity: bool = False,
    threshold: float = 0.05,
    sample_num: int = 10000,
    step: int = 50000,
    cut_height: Optional[float] = None,
    rnd_seed: int = 42,
):
    """
    Eval point to faces distance using `point_to_closest_tri_dist`.
    """
    # Setting random seeds
    random.seed(rnd_seed)
    np.random.seed(rnd_seed)

    if torch.cuda.is_available():
        dev = "cuda:0"
    else:
        dev = "cpu"
    # For testing only
    dev = "cpu"
    logger.info(f"==> [eval_single_mesh_pair] use device {dev}")

    # Load meshes
    gt_mesh = trimesh.load(gt_mesh_filename)
    pred_mesh = trimesh.load(pred_mesh_filename)
    # correct mesh gravity direction, for ADT dataset
    if correct_mesh_gravity:
        gt_mesh = correct_adt_mesh_gravity(gt_mesh)

    # cut the meshes to a plane if needed
    if cut_height is not None:
        cutting_plane = [[0, 0, -1], [0, 0, cut_height]]
        gt_mesh = gt_mesh.slice_plane(
            plane_origin=cutting_plane[1], plane_normal=cutting_plane[0]
        )
        pred_mesh = pred_mesh.slice_plane(
            plane_origin=cutting_plane[1], plane_normal=cutting_plane[0]
        )

    # convert vertices and faces to torch tensors
    pred_vertices = torch.from_numpy(pred_mesh.vertices.view(np.ndarray)).to(dev)
    gt_vertices = torch.from_numpy(gt_mesh.vertices.view(np.ndarray)).to(dev)
    pred_faces = torch.from_numpy(pred_mesh.faces.view(np.ndarray)).to(dev)
    gt_faces = torch.from_numpy(gt_mesh.faces.view(np.ndarray)).to(dev)
    logger.info(f"gt vertices and faces {gt_vertices.shape}, {gt_faces.shape}")
    logger.info(f"pred vertices and faces {pred_vertices.shape}, {pred_faces.shape}")

    # Metric 1: Compute accuracy (from sampled point in pred, to GT)
    pred_pts, _ = trimesh.sample.sample_surface(pred_mesh, sample_num, seed=rnd_seed)
    pred_pts = torch.from_numpy(pred_pts.view(np.ndarray)).to(dev)
    accuracy = compute_pts_to_mesh_dist(pred_pts, gt_faces, gt_vertices, step)

    # Metric 2: completeness
    gt_pts, _ = trimesh.sample.sample_surface(gt_mesh, sample_num, seed=rnd_seed)
    gt_pts = torch.from_numpy(gt_pts.view(np.ndarray)).to(dev)
    completeness = compute_pts_to_mesh_dist(gt_pts, pred_faces, pred_vertices, step)

    # Compute precision and recall based on accuracy and completeness
    precision_perc5 = np.mean((accuracy < 0.05).astype("float"))
    recall_perc5 = np.mean((completeness < 0.05).astype("float"))
    fscore_perc5 = 2 * precision_perc5 * recall_perc5 / (precision_perc5 + recall_perc5)

    # sort to get percentile numbers.
    metrics = {
        "Accuracy_mean_meters": np.mean(accuracy),
        "Completeness_mean_meters": np.mean(completeness),
        "prec@0.05": precision_perc5,
        "recal@0.05": recall_perc5,
        "fscore@0.05": fscore_perc5,
    }

    return metrics, accuracy, completeness


def evaluate_mesh_over_a_dataset(
    input_folder: str,
    pred_mesh_filename: str,
    gt_mesh_filename: str,
    correct_mesh_gravity: bool = False,
    threshold: float = 0.05,
    sample_num: int = 10000,
    step: int = 50000,
    cut_height: Optional[float] = None,
    rnd_seed: int = 42,
):
    logger.info(f"==> [eval_mesh_over_a_dataset] ")

    # get all the pred and gt mesh_files
    pred_mesh_paths, gt_mesh_paths = [], []
    sequence_names = os.listdir(input_folder)
    dirs = [os.path.join(input_folder, seq) for seq in sequence_names]
    dirs = [d for d in dirs if os.path.isdir(d)]
    dirs = sorted(dirs)
    for d in dirs:
        pred_mesh = os.path.join(d, pred_mesh_filename)
        gt_mesh = os.path.join(d, gt_mesh_filename)
        if os.path.exists(gt_mesh) and os.path.exists(pred_mesh):
            pred_mesh_paths.append(pred_mesh)
            gt_mesh_paths.append(gt_mesh)

    # Process each mesh pair
    overall_metrics = {
        "Accuracy_mean_meters": 0,
        "Completeness_mean_meters": 0,
        "prec@0.05": 0,
        "recal@0.05": 0,
        "fscore@0.05": 0,
    }
    for single_pred_mesh, single_gt_mesh in zip(pred_mesh_paths, gt_mesh_paths):
        logger.info(f" Evaluating over {single_pred_mesh} vs {single_gt_mesh}")
        metrics, _, _ = evaluate_single_mesh_pair(
            single_pred_mesh,
            single_gt_mesh,
            correct_mesh_gravity=correct_mesh_gravity,
            threshold=threshold,
            sample_num=sample_num,
            step=step,
            cut_height=cut_height,
            rnd_seed=rnd_seed,
        )
        for metric_key, metric_val in metrics.items():
            overall_metrics[metric_key] += metric_val

    # Aggregate metrics over all mesh pairs
    num_sequences = len(pred_mesh_paths)
    for metric_key, metric_val in overall_metrics.items():
        overall_metrics[metric_key] = metric_val / num_sequences
    return overall_metrics



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Dict, List, Optional, Tuple

import numpy as np
import torch

from atek_v1.data_preprocess.adt_gt_data_processor import AdtGtDataProcessor
from atek_v1.data_preprocess.frame_data_processor import FrameDataProcessor
from atek_v1.data_preprocess.mps_data_processor import MpsDataProcessor

from projectaria_tools.core.stream_id import StreamId
from projectaria_tools.projects.adt import AriaDigitalTwinDataPathsProvider


class AriaStreamIds:
    rgb_stream_id = StreamId("214-1")
    slaml_stream_id = StreamId("1201-1")
    slamr_stream_id = StreamId("1201-2")


class AtekRawFrameDataset(torch.utils.data.Dataset):
    """
    Dataset class to access raw ATEK data (ADT/ASE).

    This dataset class uses the data processor to convert the original ADT/ASE data format (VRS)
    into sequence of Frame's (as in data_preprocess/data_schema.py). Each Frame contains a RGB,
    SLAM-L, or SLAM-R image with object annotations and other metadata. Additionally, this dataset
    can take some Callable to convert Frame into another format, e.g., dict, for downstream use.

    Args:
        data_path (str): root directory for a sequence.
        selected_device_number (int): device number to use, there could be multiple device's
            (e.g. N) recordings in the same dataset (sequence). This number indicate which device's
            recording is chosen to process (0 to N-1).
        stream_id (AriaStreamIds): stream id to use, rgb_stream_id, slaml_stream_id, or
            slamr_stream_id.
        rotate_image_cw90deg (bool): whether to rotate image clockwise by 90 degrees.
        target_image_resolution (List[int]): image resolution (width, height) after processing.
        transform_fn (Callable): a Callable to convert Frame into another format
    """

    def __init__(
        self,
        data_path: str,
        selected_device_number: int = 0,
        stream_id: AriaStreamIds = AriaStreamIds.rgb_stream_id,
        rotate_image_cw90deg: bool = True,
        target_image_resolution: Tuple[int] = (512, 512),
        transform_fn: Optional[Callable] = None,
    ):
        paths_provider = AriaDigitalTwinDataPathsProvider(data_path)
        data_paths = paths_provider.get_datapaths_by_device_num(selected_device_number)

        mps_data_processor = MpsDataProcessor(
            name="mps",
            trajectory_file=data_paths.aria_trajectory_filepath,
        )
        try:
            adt_gt_data_processor = AdtGtDataProcessor(
                "adt_gt", "ADT", stream_id, data_paths
            )
        except Exception as e:
            adt_gt_data_processor = None
            print(f"Ground-truth NOT available for sequence {data_path}. {e}")
        data_processor = FrameDataProcessor(
            video_vrs=data_paths.aria_vrs_filepath,
            stream_id=stream_id,
            rotate_image_cw90deg=rotate_image_cw90deg,
            target_linear_camera_params=np.array(
                [target_image_resolution[0], target_image_resolution[1]]
            ),
            mps_data_processor=mps_data_processor,
            gt_data_processor=adt_gt_data_processor,
        )

        self.data_processor = data_processor
        self.transform_fn = transform_fn

    def __len__(self) -> int:
        return len(self.data_processor)

    def __getitem__(self, index: int) -> List[Dict]:
        image_frame = self.data_processor.get_frame_by_index(index)

        if self.transform_fn is None:
            return image_frame
        else:
            return self.transform_fn(image_frame)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Callable, List, Optional, Tuple

import numpy as np

import torch

from atek.data_preprocess.atek_data_sample import MultiFrameCameraData

from omegaconf.omegaconf import DictConfig
from projectaria_tools.core import calibration, data_provider
from projectaria_tools.core.calibration import CameraCalibration
from projectaria_tools.core.sensor_data import TimeDomain, TimeQueryOptions  # @manual
from projectaria_tools.core.stream_id import StreamId
from torchvision.transforms import InterpolationMode, v2

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class DepthImageProcessor:
    def __init__(
        self,
        depth_vrs: str,
        image_transform: Callable,
        depth_camera_calib: CameraCalibration,
        depth_camera_label: str,
        conf: DictConfig,
    ):
        # Parse in conf
        self.conf = conf
        self.image_transform = image_transform
        self.depth_camera_label = depth_camera_label
        self.depth_camera_calib = depth_camera_calib

        self.convert_zdepth_to_distance_flag = (
            "convert_zdepth_to_distance" in self.conf
            and self.conf.convert_zdepth_to_distance is True
        )

        # setting up vrs data provider
        self.depth_vrs = depth_vrs
        self.data_provider, self.stream_id = self.setup_vrs_data_provider()

        # Cache capture timestamps
        self.time_domain = getattr(TimeDomain, conf.time_domain)
        self.camera_timestamps = self.data_provider.get_timestamps_ns(
            self.stream_id, self.time_domain
        )

        # Cache the unprojected rays from every pixel location in the image
        if self.convert_zdepth_to_distance_flag:
            self.cached_unprojected_ray_norm = self._unproject_and_cache_pixels_to_rays(
                H=self.depth_camera_calib.get_image_size()[1],
                W=self.depth_camera_calib.get_image_size()[0],
            )

    def setup_vrs_data_provider(self):
        """
        Setup the vrs data provider, and only activate the stream id specified by camera_label.
        Returns: vrs_data_provider, stream_id
        """
        provider = data_provider.create_vrs_data_provider(self.depth_vrs)
        assert (
            provider is not None
        ), f"Cannot open depth vrs under path [{self.depth_vrs}]"

        # Find depth stream in provider
        result_stream_id = None
        # If conf specified `depth_stream_id`, use it directly
        if "depth_stream_id" in self.conf:
            result_stream_id = StreamId(self.conf.depth_stream_id)
            assert (
                result_stream_id in provider.get_all_streams()
            ), f"Specified stream id {result_stream_id} is not in vrs {self.depth_vrs}'s stream id list: {provider.get_all_streams()}"
        # If conf specified `depth_stream_type_id`, check if any of the existing streams match the type id, as $TYPE_ID-*
        elif "depth_stream_type_id" in self.conf:
            for stream_id in provider.get_all_streams():
                stream_numeric_name = str(stream_id)
                type_numeric_name = stream_numeric_name.split("-")[
                    0
                ]  # obtain the "214" out of "214-1"
                if type_numeric_name == self.conf.depth_stream_type_id:
                    result_stream_id = stream_id

                    break
            # check valid stream id is found
            assert (
                result_stream_id is not None
            ), f"Specified stream type id {self.conf.depth_stream_type_id} is not in vrs {self.depth_vrs}'s stream id list: {provider.get_all_streams()}"
        # If none of the above
        else:
            raise ValueError("No depth stream id or type id is specified in config")

        return provider, result_stream_id

    def _unproject_and_cache_pixels_to_rays(self, H: int, W: int):
        """
        A helper function to unproject all pixels in an image to rays, store the norm, and cache the results so that we don't need to recompute for every frame.
        """
        logger.info("Caching unprojected rays from all pixels in the image")
        unprojected_ray_image = torch.zeros((H, W), dtype=torch.float32)

        # height: [0 to H-1], width: [0 to W-1]
        height_coor, width_coor = torch.meshgrid(torch.arange(H), torch.arange(W))
        # all_pixel_coords: [H*W, 2]
        all_pixel_coords = torch.stack(
            [height_coor.reshape(-1), width_coor.reshape(-1)], dim=-1
        )
        for i in range(all_pixel_coords.shape[0]):
            single_pixel = all_pixel_coords[i].numpy()  # (h, w)
            unprojected_ray = torch.from_numpy(
                self.depth_camera_calib.unproject_no_checks(single_pixel)
            )
            unprojected_ray_image[single_pixel[0], single_pixel[1]] = torch.linalg.norm(
                unprojected_ray
            )

        logger.info("Completed computing unprojected rays")
        return unprojected_ray_image

    def _convert_from_zdepth_to_distance(
        self, z_depth_images: torch.Tensor
    ):  # [num_frames, 1, H, W]
        """
        Helper function to convert z-depth to distance (to camera). units are kept. (I think default is mm)
        """
        num_frames, C, H, W = (
            z_depth_images.shape[0],
            z_depth_images.shape[1],
            z_depth_images.shape[2],
            z_depth_images.shape[3],
        )
        assert (
            C == 1
        ), f"Only support single channel depth image, got {C} channels instead"
        distance_images = torch.zeros_like(z_depth_images, dtype=torch.float32)

        # Loop over all frames
        for i_frame in range(num_frames):
            # Channel is 1, no need to loop. Just loop over all pixel coordinates, and use cached unprojected rays results
            distance_images[i_frame, 0, :, :] = (
                z_depth_images[i_frame, 0, :, :] * self.cached_unprojected_ray_norm
            )

        return distance_images

    def get_depth_data_by_timestamps_ns(
        self, timestamps_ns: List[int]
    ) -> Optional[MultiFrameCameraData]:
        """
        Obtain images by timestamps. Image should be processed.
        returns: if successful, returns (image_data: Tensor [numFrames, numChannel, height, width], capture_timestamp: Tensor[numFrames], frame_id_in_stream: Tensor[numFrames])
                else returns None
        """
        image_list = []
        capture_timestamp_list = []
        frame_id_list = []
        for single_timestamp in timestamps_ns:
            index = self.data_provider.get_index_by_time_ns(
                stream_id=self.stream_id,
                time_ns=single_timestamp,
                time_domain=self.time_domain,
                time_query_options=TimeQueryOptions.CLOSEST,
            )
            image_data_and_record = self.data_provider.get_image_data_by_index(
                self.stream_id, index
            )
            frame_id = image_data_and_record[1].frame_number
            capture_timestamp = image_data_and_record[1].capture_timestamp_ns

            # Check if fetched frame is within tolerance
            if abs(capture_timestamp - single_timestamp) > self.conf.tolerance_ns:
                continue

            # Handle uint16 not supported by torch
            np_image = image_data_and_record[0].to_numpy_array()
            if np_image.dtype == np.uint16:
                np_image = np_image.astype(np.float32)

            # Convert depth units from mm to meters, if specified in config
            if "unit_scaling" in self.conf:
                np_image = np_image * self.conf.unit_scaling

            image = torch.from_numpy(np_image)
            if len(image.shape) == 2:
                # single channel image: [h,w] -> [c, h, w]
                image = torch.unsqueeze(image, dim=0)
            else:
                raise ValueError("Depth image is expected to have single channel")

            image_list.append(image)
            capture_timestamp_list.append(capture_timestamp)
            frame_id_list.append(frame_id)
        # End for single_timestamp

        # Check if at least one frame is successfully fetched
        if len(image_list) == 0:
            return None

        # Image transformations are handled by torchvision's transform functions.
        batched_depth_tensor = torch.stack(image_list, dim=0)
        batched_depth_tensor = self.image_transform(batched_depth_tensor)

        # Z-depth to distance conversion
        if self.convert_zdepth_to_distance_flag:
            batched_depth_tensor = self._convert_from_zdepth_to_distance(
                batched_depth_tensor
            )

        # properly clean output to desired dtype and shapes
        result = MultiFrameCameraData(
            images=batched_depth_tensor,
            capture_timestamps_ns=torch.tensor(
                capture_timestamp_list, dtype=torch.int64
            ),
            frame_ids=torch.tensor(frame_id_list, dtype=torch.int64),
        )

        # Fill in camera calibration information
        result.camera_label = self.depth_camera_label
        return result



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import csv
import logging

from functools import partial

from typing import Dict, Optional, Tuple

import numpy as np
from atek_v1.data_preprocess.base_gt_data_processor import BaseGtDataProcessor
from atek_v1.data_preprocess.data_schema import Frame
from atek_v1.data_preprocess.data_utils import insert_and_check
from projectaria_tools.core import calibration

from projectaria_tools.core.sophus import SE3
from projectaria_tools.core.stream_id import StreamId
from projectaria_tools.projects.adt import (
    AriaDigitalTwinDataPaths,
    AriaDigitalTwinDataProvider,
    bbox2d_to_image_coordinates,
)
from toolz import compose

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

ATEK_OTHER_CATETORY_ID: int = (
    0  # 0 is reserved for other categories in ATEK object taxonomy
)


def identity_transform(obj):
    """Helper identity transform function"""
    return obj


def sample_edge(p1: np.ndarray, p2: np.ndarray, n):
    """Helper function to sample edge points between two points in N equidistant samples"""
    return np.column_stack((np.linspace(p1[0], p2[0], n), np.linspace(p1[1], p2[1], n)))


def center_object_bb3d(bb3d_aabb: np.ndarray, T_world_bb3d: np.ndarray):
    """
    Helper function to transform the object coordinate to the object center
    and generate the new T_world_object with new object coordinate
    bb3d_aabb: [xmin, xmax, ymin, ymax, zmin, zmax]
    """
    object_dimension = bb3d_aabb[1::2] - bb3d_aabb[::2]
    t_center_in_object = (bb3d_aabb[1::2] + bb3d_aabb[::2]) / 2

    T_bb3d_object_center = SE3.from_quat_and_translation(
        1, np.array([0, 0, 0]), t_center_in_object
    )
    T_world_object_center = T_world_bb3d @ T_bb3d_object_center

    return object_dimension, T_world_object_center


def undistort_bb2d(
    bb2d: np.ndarray,
    src_calib: calibration.CameraCalibration,
    dst_calib: calibration.CameraCalibration,
    num_points_on_edge: int = 10,
):
    """bb2d is a np array of [x_min x_max y_min y_max] format"""
    bb2d_image_coordinates = bbox2d_to_image_coordinates(bb2d)
    points_on_edges = np.row_stack(
        [
            sample_edge(
                bb2d_image_coordinates[i],
                bb2d_image_coordinates[(i + 1) % 4],
                num_points_on_edge,
            )
            for i in range(4)
        ]
    )

    rect_bb2d_xs = []
    rect_bb2d_ys = []
    for points in points_on_edges:
        unprojected_bbox2d_ray = src_calib.unproject_no_checks(points)
        rect_bb2d_coord = dst_calib.project_no_checks(unprojected_bbox2d_ray)
        rect_bb2d_xs.append(rect_bb2d_coord[0])
        rect_bb2d_ys.append(rect_bb2d_coord[1])

    dst_image_width, dst_image_height = dst_calib.get_image_size()
    rect_bb2d_min_x = min(max(min(rect_bb2d_xs), 0.0), dst_image_width - 1.0)
    rect_bb2d_max_x = min(max(max(rect_bb2d_xs), 0.0), dst_image_width - 1.0)
    rect_bb2d_min_y = min(max(min(rect_bb2d_ys), 0.0), dst_image_height - 1.0)
    rect_bb2d_max_y = min(max(max(rect_bb2d_ys), 0.0), dst_image_height - 1.0)

    rect_bb2d = np.array(
        [
            rect_bb2d_min_x,
            rect_bb2d_max_x,
            rect_bb2d_min_y,
            rect_bb2d_max_y,
        ]
    )

    return rect_bb2d


def rotate_bb2d_cw90(
    bb2d: np.ndarray,
    image_height: int,
):
    new_x_min = image_height - 1.0 - bb2d[3]
    new_x_max = image_height - 1.0 - bb2d[2]
    new_y_min = bb2d[0]
    new_y_max = bb2d[1]
    return np.array([new_x_min, new_x_max, new_y_min, new_y_max])


class AdtGtDataProcessor(BaseGtDataProcessor):
    def __init__(
        self,
        name: str,
        data_source: str,
        stream_id: StreamId,
        data_path: AriaDigitalTwinDataPaths,
        category_mapping: Optional[Dict[str, Tuple[str, int]]] = None,
        instance_field_to_map: Optional[str] = None,
    ):
        """
        Initialize the AdtGtDataProcessor object.

        Args:
            name (str): The name of the data processor.
            data_source (str): The source of the data.
            stream_id (StreamId): The stream ID to process:
                "214-1" for RGB
                "1201-1" for SlamL
                "1201-2" for SlamR
            data_path (AriaDigitalTwinDataPaths): a ADTDataPath object that points to the ADT data's subtour.
            category_mapping (Optional[Dict]): A dict that maps from original data to desired taxonomy: {$instance_field_to_map: (atek_category_name, atek_category_id)}. If None, category will not be remapped.
            instance_field_to_map: (Optional[str]): The field in ADT instance that category mapping will map from, choose from: {"prototype", "category"}
        """
        super().__init__(name, stream_id)
        self.data_source = data_source
        self.gt_provider = AriaDigitalTwinDataProvider(data_path)
        self.bb2d_transform_fn = identity_transform
        self.category_mapping = category_mapping
        self.instance_field_to_map = instance_field_to_map

    def set_undistortion_params(
        self,
        original_camera_calibration: calibration.CameraCalibration,
        target_camera_calibration: Optional[calibration.CameraCalibration] = None,
        rotate_image_cw90deg: bool = True,
    ):
        """
        Set the camera calibration transformation setting to the gt generator.
        The same transformation should be used to undistort the images. This will
        be applied to the 2d bounding box gt processing. Please make sure this is
        properly set before calling `get_object_gt_at_timestamp_ns`.
        """
        image_width, image_height = original_camera_calibration.get_image_size()
        if target_camera_calibration is None:
            bb2d_undistort_fn = identity_transform
        else:
            image_width, image_height = target_camera_calibration.get_image_size()
            bb2d_undistort_fn = partial(
                undistort_bb2d,
                src_calib=original_camera_calibration,
                dst_calib=target_camera_calibration,
                num_points_on_edge=10,
            )

        if rotate_image_cw90deg:
            bb2d_rotate_fn = partial(rotate_bb2d_cw90, image_height=image_height)
        else:
            bb2d_rotate_fn = identity_transform

        self.bb2d_transform_fn = compose(bb2d_rotate_fn, bb2d_undistort_fn)

    def get_object_gt_at_timestamp_ns(
        self,
        frame: Frame,
        timestamp_ns: int,
        T_camera_world: SE3,
        tolerance_ns: int = 1000_0000,
    ):
        # We expecting valid data for the adt frames. Initialize the empty object data for frame.
        # Note that there are rare cases that frame can not find the corresponding gt in the time
        # threshold. Use the empty list or dict for those cases instead of returning None for soft
        # handling.
        frame.category_id_to_name = {}
        frame.object_instance_ids = []
        frame.object_category_ids = []
        frame.Ts_world_object = []
        frame.object_dimensions = []
        frame.Ts_camera_object = []
        frame.bb2ds = []

        # Getting 2d bounding boxes from GT first. We must need to have 2d info first to
        # find the 3d bounding box visibility. Default by querying the closest timestamp.
        bbox2d_with_dt = self.gt_provider.get_object_2d_boundingboxes_by_timestamp_ns(
            timestamp_ns,
            self.stream_id,
        )
        if not bbox2d_with_dt.is_valid():
            logger.warn(
                f"Can not get good 2d bounding boxes at {timestamp_ns} because input time is invalid"
            )
            return

        if bbox2d_with_dt.dt_ns() > tolerance_ns:
            logger.warn(
                f"Can not get good 2d bounding boxes at {timestamp_ns} because "
                f"the nearest bb2d with delta time {bbox2d_with_dt.dt_ns()}ns "
                f"bigger than the threshold we have {tolerance_ns}ns "
            )
            return

        # Get 3d bounding boxes data
        bbox3d_with_dt = self.gt_provider.get_object_3d_boundingboxes_by_timestamp_ns(
            timestamp_ns,
        )
        bb3d_available = True
        if not bbox3d_with_dt.is_valid():
            logger.warn(
                f"Can not get good 3d bounding boxes at {timestamp_ns} because input time is invalid"
            )
            bb3d_available = False

        if bbox3d_with_dt.dt_ns() > tolerance_ns:
            logger.warn(
                f"Can not get good 3d bounding boxes at {timestamp_ns} because "
                f"the nearest bb2d with delta time {bbox3d_with_dt.dt_ns()}ns "
                f"bigger than the threshold we have {tolerance_ns}ns "
            )
            bb3d_available = False

        bbox2d_all = bbox2d_with_dt.data()
        for instance_id, bbox2d_data in bbox2d_all.items():
            instance_info = self.gt_provider.get_instance_info_by_id(instance_id)
            if not self.category_mapping:
                # If no category mapping is provided, we use the original category name.
                category_name = instance_info.category
                category_id = instance_info.category_uid
            else:
                # Query the mapping field from instance
                if self.instance_field_to_map == "prototype":
                    field_value_to_map = instance_info.prototype_name
                elif self.instance_field_to_map == "category":
                    field_value_to_map = instance_info.category
                else:
                    raise ValueError(
                        f'Unsupported instance field to map: {self.instance_field_to_map}, need to be ["category", "prototype"]'
                    )

                # Perform mapping
                if field_value_to_map in self.category_mapping:
                    category_name = self.category_mapping[field_value_to_map][0]
                    category_id = self.category_mapping[field_value_to_map][1]
                else:
                    category_name = "other"
                    category_id = ATEK_OTHER_CATETORY_ID

            insert_and_check(frame.category_id_to_name, category_id, category_name)
            frame.object_instance_ids.append(instance_id)
            frame.object_category_ids.append(category_id)
            frame.bb2ds.append(self.bb2d_transform_fn(bbox2d_data.box_range))

            if bb3d_available:
                bbox3d_data = bbox3d_with_dt.data()[instance_id]
                bb3d_aabb = bbox3d_data.aabb
                T_world_bb3d = bbox3d_data.transform_scene_object
                object_dimension, T_world_object = center_object_bb3d(
                    bb3d_aabb, T_world_bb3d
                )

                frame.object_dimensions.append(object_dimension)
                frame.Ts_world_object.append(T_world_object.to_matrix3x4())
                frame.Ts_camera_object.append(
                    (T_camera_world @ T_world_object).to_matrix3x4()
                )

        assert (
            len(frame.object_instance_ids)
            == len(frame.object_category_ids)
            == len(frame.bb2ds)
        )
        if bb3d_available:
            assert (
                len(frame.object_category_ids)
                == len(frame.object_dimensions)
                == len(frame.Ts_world_object)
                == len(frame.Ts_camera_object)
            ), (
                f"{len(frame.object_category_ids)}, {len(frame.object_dimensions)}, "
                f"{len(frame.Ts_world_object)}, {len(frame.Ts_camera_object)}"
            )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Union

import numpy as np
import seaborn as sns

import torch
from atek_v1.data_preprocess.data_schema import Frameset, FramesetGroup


def get_rate_stats(timestamps_ns: List[int]):
    deltas_ns = np.diff(timestamps_ns)
    deltas_ms = deltas_ns / 1000_000
    framerate_hz = 1 / np.mean(deltas_ms / 1000)
    mean_delta_ms = np.mean(deltas_ms)
    std_delta_ms = np.std(deltas_ms)
    max_delta_ms = np.max(deltas_ms)
    min_delta_ms = np.min(deltas_ms)
    return {
        "rate_hz": framerate_hz,
        "mean_delta_ms": mean_delta_ms,
        "std_delta_ms": std_delta_ms,
        "max_delta_ms": max_delta_ms,
        "min_delta_ms": min_delta_ms,
    }


def insert_and_check(my_dict, key, value):
    """
    Helper function to insert a new key-value pair into a dictionary,
    or check that the existing value matches the new one
    """
    if key not in my_dict:
        my_dict[key] = value
    elif my_dict[key] == value:
        pass
    else:
        raise AssertionError(f"Key {key} exists with a different value {my_dict[key]}")


def strict_update_dict(dict_A, dict_B):
    """
    Helper function to update a dictionary A with values from another dictionary B.
    The update will either insert a new key-value pair into a dictionary,
    or check that the existing value matches the new one
    """
    for k, v in dict_B.items():
        insert_and_check(dict_A, k, v)


def unify_list(input_list: List[int]):
    """
    Helper function to convert a list of integers to a list of unique values and their corresponding indices.
    """
    input_array = np.array(input_list)
    unique_values, indices = np.unique(input_array, return_index=True)

    unique_values = [int(val) for val in unique_values]
    indices = [int(idx) for idx in indices]
    return unique_values, indices


def check_all_same_member(values, member: str):
    # Extract the specified member from each object in the list
    member_values = [getattr(obj, member) for obj in values]
    # Check if all the extracted values are the same
    return len(set(member_values)) == 1


def generate_disjoint_colors(n):
    palette = sns.color_palette("husl", n)
    return [tuple(int(255 * x) for x in color) for color in palette]


def unify_object_target(
    unified_target: Union[FramesetGroup, Frameset],
):
    """
    Helper function to union the multiple frames object targets into
    the frameset level object target.
    """
    if isinstance(unified_target, FramesetGroup):
        member_name = "framesets"
    elif isinstance(unified_target, Frameset):
        member_name = "frames"
    else:
        raise ValueError("Invalid type for the unified target")

    unified_target.category_id_to_name = {}
    unified_target.object_instance_ids = []
    unified_target.object_category_ids = []
    unified_target.Ts_world_object = []
    unified_target.object_dimensions = []
    for member in getattr(unified_target, member_name):
        strict_update_dict(
            unified_target.category_id_to_name, member.category_id_to_name
        )
        unified_target.object_instance_ids += member.object_instance_ids
        unified_target.object_category_ids += member.object_category_ids
        unified_target.Ts_world_object += member.Ts_world_object
        unified_target.object_dimensions += member.object_dimensions

    assert (
        len(unified_target.object_instance_ids)
        == len(unified_target.object_category_ids)
        == len(unified_target.Ts_world_object)
        == len(unified_target.object_dimensions)
    ), (
        f"{len(unified_target.object_instance_ids)}, {len(unified_target.object_category_ids)}, "
        f"{len(unified_target.Ts_world_object)}, {len(unified_target.object_dimensions)}"
    )

    # Remove the duplicate object
    unified_target.object_instance_ids, unify_ids = unify_list(
        unified_target.object_instance_ids
    )
    unified_target.object_category_ids = [
        unified_target.object_category_ids[id] for id in unify_ids
    ]
    unified_target.Ts_world_object = [
        unified_target.Ts_world_object[id] for id in unify_ids
    ]
    unified_target.object_dimensions = [
        unified_target.object_dimensions[id] for id in unify_ids
    ]


def value_to_tensor(value: Union[np.ndarray, List[np.ndarray]]) -> torch.Tensor:
    """
    Convert a numpy array or a list of same shape numpy array tensor to a tensor.
    """
    result = None
    if isinstance(value, np.ndarray):
        result = torch.from_numpy(value)
    elif isinstance(value, list):
        result = (
            torch.stack([torch.from_numpy(item) for item in value])
            if len(value) > 0
            else None
        )
    else:
        raise ValueError(f"Unsupported type {type(value)}")

    return result



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List


class CubercnnPredictionConverter:
    """
    Convert CubeRCNN model predictions from detectron2 instance to a list of dicts

    Args:
        score_threshold (float): threshold to filter out low-confidence predictions
        category_names (List[str]): the list of predicted object categories.
    """

    def __init__(self, score_threshold: str, category_names: List[str]):
        self.score_threshold = score_threshold
        self.category_names = category_names

    def __call__(self, model_input: List[Dict], model_prediction: List[Dict]):
        """
        Converts per-frame CubeRCNN prediction to list of dicts
        """
        converted_model_predictions = []

        for input, prediction in zip(model_input, model_prediction):
            dets = prediction["instances"]
            preds_per_frame = []

            if len(dets) == 0:
                continue

            for (
                corners3D,
                center_cam,
                center_2D,
                dimensions,
                bbox_2D,
                pose,
                score,
                scores_full,
                cat_idx,
            ) in zip(
                dets.pred_bbox3D,
                dets.pred_center_cam,
                dets.pred_center_2D,
                dets.pred_dimensions,
                dets.pred_boxes,
                dets.pred_pose,
                dets.scores,
                dets.scores_full,
                dets.pred_classes,
            ):
                if score < self.score_threshold:
                    continue
                cat = self.category_names[cat_idx]
                predictions_dict = {
                    "sequence_name": input["sequence_name"],
                    "frame_id": input["frame_id"],
                    "timestamp_ns": input["timestamp_ns"],
                    "T_world_cam": input["T_world_camera"],
                    # CubeRCNN dimensions are in reversed order of Aria data convention
                    "dimensions": dimensions.tolist()[::-1],
                    "t_cam_obj": center_cam.tolist(),
                    "R_cam_obj": pose.tolist(),
                    "corners3D": corners3D.tolist(),
                    "center_2D": center_2D.tolist(),
                    "bbox_2D": bbox_2D.tolist(),
                    "score": score.detach().item(),
                    "scores_full": scores_full.tolist(),
                    "category_idx": cat_idx.detach().item(),
                    "category": cat,
                }
                preds_per_frame.append(predictions_dict)

            converted_model_predictions.append(preds_per_frame)

        return converted_model_predictions



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional

import numpy as np

import torch

import webdataset as wds

from atek.data_loaders.atek_wds_dataloader import load_atek_wds_dataset
from atek.data_loaders.detectron2_structures.detectron2_structures import (
    Boxes,
    Instances,
)
from atek.util.atek_constants import ATEK_CATEGORY_ID_TO_NAME, ATEK_CATEGORY_NAME_TO_ID
from projectaria_tools.core.sophus import SE3

from webdataset.filters import pipelinefilter


class CubeRCNNModelAdaptor:
    def __init__(
        self,
        # TODO: make these a DictConfig
        min_bb2d_area: Optional[float] = 100,
        min_bb3d_depth: Optional[float] = 0.3,
        max_bb3d_depth: Optional[float] = 5.0,
    ):
        self.min_bb2d_area = min_bb2d_area
        self.min_bb3d_depth = min_bb3d_depth
        self.max_bb3d_depth = max_bb3d_depth

    @staticmethod
    def get_dict_key_mapping_all():
        dict_key_mapping = {
            "mfcd#camera-rgb+images": "image",
            "mfcd#camera-rgb+projection_params": "camera_params",
            "mfcd#camera-rgb+camera_model_name": "camera_model",
            "mfcd#camera-rgb+t_device_camera": "t_device_rgbcam",
            "mfcd#camera-rgb+frame_ids": "frame_id",
            "mfcd#camera-rgb+capture_timestamps_ns": "timestamp_ns",
            "mtd#ts_world_device": "ts_world_device",
            "sequence_name": "sequence_name",
            "gt_data": "gt_data",
        }
        return dict_key_mapping

    def atek_to_cubercnn(self, data):
        """
        A helper data transform function to convert a ATEK webdataset data sample built by CubeRCNNSampleBuilder, to CubeRCNN unbatched
        samples. Yield one unbatched sample a time to use the collation and batching mechanism in
        the webdataset properly.
        """
        for atek_wds_sample in data:
            sample = {}
            self._update_camera_data_in_sample(atek_wds_sample, sample)
            self._update_T_world_camera(atek_wds_sample, sample)

            # Skip if no gt data
            if "gt_data" in atek_wds_sample and len(atek_wds_sample["gt_data"]) > 0:
                self._update_gt_data_in_sample(atek_wds_sample, sample)

            yield sample

    def _update_camera_data_in_sample(self, atek_wds_sample, sample):
        """
        Initialize sample image
        Process camera K-matrix information and update the sample dictionary.
        """
        assert atek_wds_sample["image"].shape[0] == 1, "Only support 1 frame"
        image_height, image_width = atek_wds_sample["image"].shape[2:]

        # calculate K-matrix
        camera_model = atek_wds_sample["camera_model"]
        assert (
            camera_model == "CameraModelType.LINEAR"
        ), f"Only linear camera model supported in CubeRCNN model, this data has {camera_model} instead."
        k_matrix = torch.zeros((3, 3), dtype=torch.float32)
        params = atek_wds_sample["camera_params"]
        k_matrix[0, 0], k_matrix[1, 1] = params[0], params[1]
        k_matrix[0, 2], k_matrix[1, 2] = params[2], params[3]
        k_matrix[2, 2] = 1.0

        sample.update(
            {
                # rgb -> bgr
                "image": atek_wds_sample["image"][0, [2, 1, 0], :, :].clone().detach(),
                "K": k_matrix.tolist(),
                "height": image_height,
                "width": image_width,
                "K_matrix": k_matrix,
                "timestamp_ns": atek_wds_sample["timestamp_ns"],
                "frame_id": atek_wds_sample["frame_id"],
                "sequence_name": atek_wds_sample["sequence_name"],
            }
        )

    def _update_T_world_camera(self, atek_wds_sample, sample):
        """
        Compute world-to-camera transformation matrices, and update this field in sample dict.
        """
        T_world_device = SE3.from_matrix3x4(atek_wds_sample["ts_world_device"][0])
        T_device_rgbCam = SE3.from_matrix3x4(atek_wds_sample["t_device_rgbcam"])
        T_world_rgbCam = T_world_device @ T_device_rgbCam
        sample["T_world_camera"] = T_world_rgbCam.to_matrix3x4()

    def _process_2d_bbox_dict(self, bb2d_dict):
        """
        Process 2D bounding boxes by rearranging the bounding box coordinates to be
        in the order x0, y0, x1, y1 and calculating the area of each 2D bounding box.
        """
        bb2ds_x0y0x1y1 = bb2d_dict["box_ranges"]
        bb2ds_x0y0x1y1 = bb2ds_x0y0x1y1[:, [0, 2, 1, 3]]
        bb2ds_area = (bb2ds_x0y0x1y1[:, 2] - bb2ds_x0y0x1y1[:, 0]) * (
            bb2ds_x0y0x1y1[:, 3] - bb2ds_x0y0x1y1[:, 1]
        )

        return bb2ds_x0y0x1y1, bb2ds_area

    def _process_3d_bbox_dict(self, bbox3d_dict, T_world_rgbCam):
        """
        This function processes 3D bounding box data from a given dictionary,
        extracting dimensions, calculating depths, and computing transformation
        matrices relative to the camera.
        """
        bb3d_dimensions = bbox3d_dict["object_dimensions"]

        bb3d_depths_list = []
        Ts_world_object_list = []
        Ts_cam_object_list = []
        for _, pose_as_tensor in enumerate(bbox3d_dict["ts_world_object"]):
            T_world_object = SE3.from_matrix3x4(pose_as_tensor.numpy())
            T_cam_object = T_world_rgbCam.inverse() @ T_world_object

            # Add to lists
            Ts_world_object_list.append(
                torch.tensor(T_world_object.to_matrix3x4(), dtype=torch.float32)
            )

            Ts_cam_object_list.append(
                torch.tensor(T_cam_object.to_matrix3x4(), dtype=torch.float32)
            )
            bb3d_depths_list.append(T_cam_object.translation()[:, 2].item())

        # Convert lists to tensors
        bb3d_depths = torch.tensor(bb3d_depths_list, dtype=torch.float32)
        Ts_world_object = torch.stack(Ts_world_object_list, dim=0)
        Ts_cam_object = torch.stack(Ts_cam_object_list, dim=0)

        return bb3d_dimensions, bb3d_depths, Ts_world_object, Ts_cam_object

    def _update_gt_data_in_sample(self, atek_wds_sample, sample):
        """
        updates the sample dictionary with filtered ground truth data for both 2D and 3D bounding boxes.
        """
        bbox2d_dict = atek_wds_sample["gt_data"]["obb2_gt"]["camera-rgb"]
        bbox3d_dict = atek_wds_sample["gt_data"]["obb3_gt"]["camera-rgb"]

        # Instance id between obb3 and obb2 should be the same
        assert torch.allclose(
            bbox3d_dict["instance_ids"], bbox2d_dict["instance_ids"], atol=0
        ), "instance ids in obb2 and obb3 needs to be exactly the same!"

        category_ids = bbox3d_dict["category_ids"]

        T_world_rgbCam = SE3.from_matrix3x4(sample["T_world_camera"])

        bb2ds_x0y0x1y1, bb2ds_area = self._process_2d_bbox_dict(bbox2d_dict)
        bb3d_dimensions, bb3d_depths, Ts_world_object, Ts_cam_object = (
            self._process_3d_bbox_dict(bbox3d_dict, T_world_rgbCam)
        )

        # Filter 1: ignore category = -1, meaning "Other".
        category_id_filter = category_ids > 0  # filter out -1 category = "Other"

        # Filter 2: ignore bboxes with small area
        bb2d_area_filter = bb2ds_area > self.min_bb2d_area

        # Filter 3: ignore bboxes with small depth
        bb3d_depth_filter = (self.min_bb3d_depth <= bb3d_depths) & (
            bb3d_depths <= self.max_bb3d_depth
        )

        # Combine all filters
        final_filter = category_id_filter & bb2d_area_filter & bb3d_depth_filter

        # Apply filter to create instances
        image_height = sample["height"]
        image_width = sample["width"]
        instances = Instances((image_height, image_width))
        instances.gt_classes = category_ids[final_filter]
        instances.gt_boxes = Boxes(bb2ds_x0y0x1y1[final_filter])

        # Create 3D bboxes
        Ts_cam_object_filtered = Ts_cam_object[final_filter]
        trans_cam_object_filtered = Ts_cam_object_filtered[:, :, 3]
        k_matrix = sample["K_matrix"]
        filtered_projection_2d = (
            k_matrix.repeat(len(trans_cam_object_filtered), 1, 1)
            @ trans_cam_object_filtered.unsqueeze(-1)
        ).squeeze(-1)
        filtered_projection_2d = filtered_projection_2d[:, :2] / filtered_projection_2d[
            :, 2
        ].unsqueeze(-1)
        instances.gt_boxes3D = torch.cat(
            [
                filtered_projection_2d,  # [N, 2]
                bb3d_depths[final_filter].unsqueeze(-1).clone().detach(),  # [N, 1]
                # Omni3d has the inverted zyx dimensions
                # https://github.com/facebookresearch/omni3d/blob/main/cubercnn/util/math_util.py#L144C1-L181C40
                bb3d_dimensions[final_filter].flip(-1).clone().detach(),  # [N, 3]
                trans_cam_object_filtered,  # [N, 3]
            ],
            axis=-1,
        )
        instances.gt_poses = Ts_cam_object_filtered[:, :, :3].clone().detach()

        # Update sample with filtered instance data
        sample["instances"] = instances
        sample["Ts_world_object"] = Ts_world_object[final_filter].clone().detach()
        sample["object_dimensions"] = bb3d_dimensions[final_filter].clone().detach()
        sample["category"] = category_ids[final_filter].clone().detach()

    @staticmethod
    def cubercnn_gt_to_atek_gt(
        cubercnn_dict: Dict,
        T_world_camera_np: np.array,
        camera_label: str = "camera-rgb",
        cubercnn_id_to_atek_id: Optional[Dict[int, int]] = None,
    ) -> Optional[Dict]:
        """
        A helper data transform function to convert the model input (gt) dict, or output (prediction) dict from CubeRCNN format,
        back to ATEK GT dict format (defined in `obb_sample_builder`, which is effectively obb3_gt_processor + obb2_gt_processor)
        CubeRCNN model is ran only on one camera stream, so user should specific which camera stream to use. By default, it is "camera-rgb".
        """
        cubercnn_instances = cubercnn_dict["instances"]
        # Skip if no instances
        if len(cubercnn_instances) == 0:
            return None

        # Check if the cubercnn_dict is a prediction dict or gt dict. If it is a gt dict,
        # "transfer" it to a prediction dict by filling in the pred fields. TODO: Consider another way to handle this!
        pred_flag = hasattr(cubercnn_instances, "pred_classes")
        if not pred_flag:
            # fake pred fields using gt fields
            num_instances = len(cubercnn_instances.gt_classes)
            cubercnn_instances.pred_classes = cubercnn_instances.gt_classes
            cubercnn_instances.pred_boxes = cubercnn_instances.gt_boxes
            cubercnn_instances.pred_dimensions = cubercnn_instances.gt_boxes3D[:, 3:6]
            cubercnn_instances.pred_center_cam = cubercnn_instances.gt_boxes3D[:, 6:9]
            cubercnn_instances.pred_pose = cubercnn_instances.gt_poses
            cubercnn_instances.scores = torch.ones(num_instances, dtype=torch.float32)

        # initialize ATEK GT dict
        atek_dict = {
            "obb3_gt": {},
            "obb2_gt": {},
            "scores": cubercnn_instances.scores.detach().cpu(),  # tensor, shape: [num_instances], float32
        }
        atek_dict["obb3_gt"][camera_label] = {
            "instance_ids": None,
            "category_names": None,
            "category_ids": cubercnn_instances.pred_classes.detach().cpu(),
        }
        atek_dict["obb2_gt"][camera_label] = {
            "instance_ids": None,
            "category_names": None,
            "category_ids": cubercnn_instances.pred_classes.detach().cpu(),
            "visibility_ratios": None,
        }

        # Fill in category ids
        if cubercnn_id_to_atek_id is not None:
            atek_id_list = [
                cubercnn_id_to_atek_id[id.item()]
                for id in cubercnn_instances.pred_classes
            ]
            atek_dict["obb3_gt"][camera_label]["category_ids"] = torch.tensor(
                atek_id_list, dtype=torch.int32
            )
            atek_dict["obb2_gt"][camera_label]["category_ids"] = torch.tensor(
                atek_id_list, dtype=torch.int32
            )
        else:
            atek_dict["obb3_gt"][camera_label][
                "category_ids"
            ] = cubercnn_instances.pred_classes.detach().cpu()
            atek_dict["obb2_gt"][camera_label][
                "category_ids"
            ] = cubercnn_instances.pred_classes.detach().cpu()

        # Fill category names
        atek_dict["obb3_gt"][camera_label]["category_names"] = [
            ATEK_CATEGORY_ID_TO_NAME[id.item()]
            for id in atek_dict["obb3_gt"][camera_label]["category_ids"]
        ]
        atek_dict["obb2_gt"][camera_label]["category_names"] = [
            ATEK_CATEGORY_ID_TO_NAME[id.item()]
            for id in atek_dict["obb2_gt"][camera_label]["category_ids"]
        ]

        # CubeRCNN dimensions are in reversed order (zyx) compared to ATEK (xyz)
        bbox3d_dim = (
            cubercnn_instances.pred_dimensions.detach().cpu()
        )  # tensor, shape [num_instances, 3]
        atek_dict["obb3_gt"][camera_label]["object_dimensions"] = torch.flip(
            bbox3d_dim, dims=[1]
        )

        # Fill in pose
        rotations = cubercnn_instances.pred_pose.detach().cpu()  # [num_instances, 3, 3]
        translations = (
            cubercnn_instances.pred_center_cam.detach().cpu().unsqueeze(2)
        )  # [num_instances, 3, 1]

        Ts_cam_object = SE3.from_matrix3x4(
            torch.cat((rotations, translations), dim=2).numpy()
        )
        T_world_cam = SE3.from_matrix3x4(T_world_camera_np)

        Ts_world_object = T_world_cam @ Ts_cam_object
        Ts_world_object = SE3.to_matrix3x4(Ts_world_object)  # [num_instances, 3, 4]
        if Ts_world_object.shape == (3, 4):
            Ts_world_object = Ts_world_object.reshape(1, 3, 4)
        atek_dict["obb3_gt"][camera_label]["ts_world_object"] = torch.tensor(
            Ts_world_object, dtype=torch.float32
        )

        # Fill in 2d bbox ranges
        bbox2d = (
            cubercnn_instances.pred_boxes.tensor.detach().cpu()
        )  # tensor, shape [num_instances, 4]
        # x0,y0,x1,y1 -> x0,x1,y0,y1
        atek_dict["obb2_gt"][camera_label]["box_ranges"] = torch.stack(
            (bbox2d[:, 0], bbox2d[:, 2], bbox2d[:, 1], bbox2d[:, 3]), dim=1
        )

        return atek_dict


def cubercnn_collation_fn(batch):
    # Simply collate as a list
    return list(batch)


def load_atek_wds_dataset_as_cubercnn(
    urls: List, batch_size: Optional[int], repeat_flag: bool, shuffle_flag: bool = False
) -> wds.WebDataset:
    cubercnn_model_adaptor = CubeRCNNModelAdaptor()

    return load_atek_wds_dataset(
        urls,
        batch_size=batch_size,
        dict_key_mapping=CubeRCNNModelAdaptor.get_dict_key_mapping_all(),
        data_transform_fn=pipelinefilter(cubercnn_model_adaptor.atek_to_cubercnn)(),
        collation_fn=cubercnn_collation_fn,
        repeat_flag=repeat_flag,
        shuffle_flag=shuffle_flag,
    )


def create_atek_dataloader_as_cubercnn(
    urls: List[str],
    batch_size: Optional[int] = None,
    repeat_flag: bool = False,
    shuffle_flag: bool = False,
    num_workers: int = 0,
) -> torch.utils.data.DataLoader:
    wds_dataset = load_atek_wds_dataset_as_cubercnn(
        urls,
        batch_size=batch_size,
        repeat_flag=repeat_flag,
        shuffle_flag=shuffle_flag,
    )

    return torch.utils.data.DataLoader(
        wds_dataset, batch_size=None, num_workers=num_workers, pin_memory=True
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import json
import os
import re
from dataclasses import dataclass

from typing import Callable, Dict, List, Optional

import numpy as np
import torch
import webdataset as wds

from atek_v1.data_preprocess.data_schema import Frame, Frameset, FramesetGroup
from PIL import Image


def tensorize_value(x):
    """
    Helper function to recursively tensorize the input object. Only expect list nested input.
    This function will convert the value to a single tensor.
    Return None for empty list.
    """
    if isinstance(x, torch.Tensor):
        return x
    elif isinstance(x, list):
        if len(x) == 0:
            return None
        return torch.stack([tensorize_value(xi) for xi in x], dim=0)
    elif isinstance(x, np.ndarray):
        return torch.from_numpy(x)
    else:
        raise ValueError(f"Unsupported type {type(x)}")


def get_info_json(
    flatten_fsg_dict: Dict,
    key_pattern: re.Pattern,
    selected_fields: List,
):
    """
    A helper function to generate a json string from selected dict values based on keys.
    This is better suitable for metadata.
    """
    info_json_dict = {}
    for k, v in flatten_fsg_dict.items():
        if key_pattern.search(k) and k.split("+")[-1] in selected_fields:
            info_json_dict[k] = v
    return info_json_dict


def get_tensor_dict(
    flatten_fsg_dict: Dict,
    key_pattern: re.Pattern,
    selected_fields: List,
    return_as_list_of_tensor: bool = False,
):
    """
    A helper function to generate a dict tensor from selected dict values based on keys.
    This is better suitable for tensor style of data.

    return_as_list_of_tensor: by default set to false so that try the best to convert values
    into a single tensor. However, some data is not aligned in length for example objects, so we
    return a list of tensors instead by setting return_as_list_of_tensor to true.
    """
    tensor_dict = {}
    for k, v in flatten_fsg_dict.items():
        if key_pattern.search(k) and k.split("+")[-1] in selected_fields:
            if return_as_list_of_tensor:
                tensor_dict[k] = [tensorize_value(vi) for vi in v]
            else:
                tensor_dict[k] = tensorize_value(v)
    return tensor_dict


def add_frame_images_to_wds(flatten_fsg_dict, wds_dict):
    frame_key_pattern = Frame.flatten_dict_key_pattern()
    for k, v in flatten_fsg_dict.items():
        if frame_key_pattern.search(k) and k.split("+")[-1] in Frame.image_fields():
            for i, image in enumerate(v):
                wds_dict[f"{k}_{i}.jpeg"] = image


def add_frame_camera_info_to_wds(flatten_fsg_dict, wds_dict):
    wds_dict["frame_camera_info.json"] = get_info_json(
        flatten_fsg_dict, Frame.flatten_dict_key_pattern(), Frame.camera_info_fields()
    )


def add_frame_camera_data_to_wds(flatten_fsg_dict, wds_dict):
    wds_dict["frame_camera_data.pth"] = get_tensor_dict(
        flatten_fsg_dict, Frame.flatten_dict_key_pattern(), Frame.camera_data_fields()
    )


def add_frame_info_to_wds(flatten_fsg_dict, wds_dict):
    wds_dict["frame_info.json"] = get_info_json(
        flatten_fsg_dict, Frame.flatten_dict_key_pattern(), Frame.input_info_fields()
    )


def add_frame_trajectory_data_to_wds(flatten_dict, wds_dict):
    wds_dict["frame_trajectory_data.pth"] = get_tensor_dict(
        flatten_dict,
        Frame.flatten_dict_key_pattern(),
        Frame.trajectory_data_fields(),
    )


def add_frame_object_info_to_wds(flatten_fsg_dict, wds_dict):
    wds_dict["frame_object_info.json"] = get_info_json(
        flatten_fsg_dict,
        Frame.flatten_dict_key_pattern(),
        Frame.object_info_fields(),
    )


def add_frame_object_2d_data_to_wds(flatten_fsg_dict, wds_dict):
    wds_dict["frame_object_2d.pth"] = get_tensor_dict(
        flatten_fsg_dict,
        Frame.flatten_dict_key_pattern(),
        Frame.object_2d_data_fields(),
        return_as_list_of_tensor=True,
    )


def add_frame_object_3d_data_to_wds(flatten_fsg_dict, wds_dict):
    wds_dict["frame_object_3d.pth"] = get_tensor_dict(
        flatten_fsg_dict,
        Frame.flatten_dict_key_pattern(),
        Frame.object_3d_data_fields(),
        return_as_list_of_tensor=True,
    )


def add_frameset_info_to_wds(flatten_fsg_dict, wds_dict):
    wds_dict["frameset_info.json"] = get_info_json(
        flatten_fsg_dict,
        Frameset.flatten_dict_key_pattern(),
        Frameset.input_info_fields(),
    )


def add_frameset_camera_data_to_wds(flatten_fsg_dict, wds_dict):
    wds_dict["frameset_camera_data.pth"] = get_tensor_dict(
        flatten_fsg_dict,
        Frameset.flatten_dict_key_pattern(),
        Frameset.camera_data_fields(),
    )


def add_frameset_trajectory_data_to_wds(flatten_fsg_dict, wds_dict):
    wds_dict["frameset_trajectory_data.pth"] = get_tensor_dict(
        flatten_fsg_dict,
        Frameset.flatten_dict_key_pattern(),
        Frameset.trajectory_data_fields(),
    )


def add_frameset_semidense_points_to_wds(flatten_dict, wds_dict):
    wds_dict["frameset_semidense_points_data.pth"] = get_tensor_dict(
        flatten_dict,
        Frameset.flatten_dict_key_pattern(),
        Frameset.semidense_points_fields(),
        return_as_list_of_tensor=True,
    )


def add_frameset_object_info_to_wds(flatten_fsg_dict, wds_dict):
    wds_dict["frameset_object_info.json"] = get_info_json(
        flatten_fsg_dict,
        Frameset.flatten_dict_key_pattern(),
        Frameset.object_info_fields(),
    )


def add_frameset_object_3d_data_to_wds(flatten_fsg_dict, wds_dict):
    wds_dict["frameset_object_3d.pth"] = get_tensor_dict(
        flatten_fsg_dict,
        Frameset.flatten_dict_key_pattern(),
        Frameset.object_3d_data_fields(),
        return_as_list_of_tensor=True,
    )


def add_frameset_group_info_to_wds(flatten_fsg_dict, wds_dict):
    wds_dict["frameset_group_info.json"] = get_info_json(
        flatten_fsg_dict,
        FramesetGroup.flatten_dict_key_pattern(),
        FramesetGroup.input_info_fields(),
    )


def add_frameset_group_trajectory_data_to_wds(flatten_fsg_dict, wds_dict):
    wds_dict["frameset_group_trajectory_data.pth"] = get_tensor_dict(
        flatten_fsg_dict,
        FramesetGroup.flatten_dict_key_pattern(),
        FramesetGroup.trajectory_data_fields(),
    )


def add_frameset_group_object_info_to_wds(flatten_fsg_dict, wds_dict):
    wds_dict["frameset_group_object_info.json"] = get_info_json(
        flatten_fsg_dict,
        FramesetGroup.flatten_dict_key_pattern(),
        FramesetGroup.object_info_fields(),
    )


def add_frameset_group_object_3d_data_to_wds(flatten_fsg_dict, wds_dict):
    wds_dict["frameset_group_object_3d.pth"] = get_tensor_dict(
        flatten_fsg_dict,
        FramesetGroup.flatten_dict_key_pattern(),
        FramesetGroup.object_3d_data_fields(),
    )


def encrypt_wds_dict(wds_dict, encryption_handler: Callable[[bytes], bytes]):
    """
    Encodes the given WDS dictionary based on the keys' file extension to different
    files. The dictionary is expected to have keys that end with .jpeg, .json, or .pth,
    and the corresponding values should be in a format that can be serialized into
    these formats.
    """
    encoded_dict = {}
    for k, v in wds_dict.items():
        if k == "__key__":
            encoded_dict[k] = v
        elif k.endswith(".jpeg"):
            # Assuming 'v' is a numpy array representing an image
            # Convert to PIL Image and then to bytes
            pil_image = Image.fromarray(v)
            with io.BytesIO() as output_bytes:
                pil_image.save(output_bytes, format="JPEG")
                encrypted_val = encryption_handler(output_bytes.getvalue())
            encoded_dict[k] = encrypted_val
        elif k.endswith(".json"):
            # Assuming 'v' is serializable to JSON
            encoded_dict[k] = encryption_handler(json.dumps(v).encode("utf-8"))
        elif k.endswith(".pth"):
            # Assuming 'v' is a PyTorch tensor or a dictionary of tensors
            with io.BytesIO() as buffer:
                torch.save(v, buffer)
                buffer.seek(0)
                encrypted_val = encryption_handler(buffer.read())
            encoded_dict[k] = encrypted_val
        else:
            raise ValueError(f"Unknown key extension {k}")

    return encoded_dict


@dataclass
class DataSelectionSettings:
    # Flag to include per-frame trajectory information.
    require_traj_for_frame: bool = False

    # Flag to include per-frame 2D object bounding boxes.
    require_obb2d_gt_for_frame: bool = False

    # Flag to include per-frame 3D object bounding boxes.
    require_obb3d_gt_for_frame: bool = False

    # Flag to include per-frameset trajectory information.
    require_traj_for_frameset: bool = False

    # Flag to include per-frameset semidense point cloud information
    require_semidense_points_for_frameset: bool = False

    # Flag to include per-frameset 3D object bounding boxes.
    require_obb3d_gt_for_frameset: bool = False

    # Flag to include overall frameset group trajectory information.
    require_traj_for_frameset_group: bool = False

    # Flag to include overall frameset group 3D object bounding boxes.
    require_obb3d_gt_for_frameset_group: bool = False


def convert_frameset_group_to_wds_dict(
    wds_id: int,
    frameset_group: FramesetGroup,
    wds_data_prefix_string: str,
    data_selection_settings: DataSelectionSettings,
):
    """
    Prepares a FramesetGroup object for serialization with WebDataset by flattening
    its structure and collating required fields into appropriate formats. The output is
    organized into multiple files, such as images, metadata JSONs, and tensor .pth files.

    Primitive types like ints and strs are stored in lists or dicts for clarity and
    readability as metadata. Numerical data like poses and bounding boxes are tensorized,
    while images are kept as NumPy arrays to facilitate easy image IO operations.

    The function selectively includes various pieces of trajectory and bounding box
    information based on specified requirements. This tailored serialization supports
    flexibility and efficiency in dataset storage and retrieval.

    Args:
        wds_id: Unique ID for the WDS sample.
        frameset_group: The FramesetGroup object to serialize.
        data_selection_settings: Settings for selecting data from the FramesetGroup.
        wds_data_prefix_string: Prefix string to prepend to `__key__`, which should be unique across datasets because WDS has assumptions that filenames across different tars need to be unique
    """
    flatten_fsg_dict = frameset_group.to_flatten_dict()

    wds_dict = {"__key__": f"{wds_data_prefix_string}_fsg_{wds_id:06}"}

    # Default fields which we want to keep always.
    add_frame_info_to_wds(flatten_fsg_dict, wds_dict)
    add_frame_images_to_wds(flatten_fsg_dict, wds_dict)
    add_frame_camera_info_to_wds(flatten_fsg_dict, wds_dict)
    add_frame_camera_data_to_wds(flatten_fsg_dict, wds_dict)
    add_frameset_info_to_wds(flatten_fsg_dict, wds_dict)
    add_frameset_camera_data_to_wds(flatten_fsg_dict, wds_dict)
    add_frameset_group_info_to_wds(flatten_fsg_dict, wds_dict)

    # Trajectory data.
    if data_selection_settings.require_traj_for_frame:
        add_frame_trajectory_data_to_wds(flatten_fsg_dict, wds_dict)
    if data_selection_settings.require_traj_for_frameset:
        add_frameset_trajectory_data_to_wds(flatten_fsg_dict, wds_dict)
    if data_selection_settings.require_traj_for_frameset_group:
        add_frameset_group_trajectory_data_to_wds(flatten_fsg_dict, wds_dict)

    # Semidense point cloud data.
    if data_selection_settings.require_semidense_points_for_frameset:
        add_frameset_semidense_points_to_wds(flatten_fsg_dict, wds_dict)

    # Object data
    if (
        data_selection_settings.require_obb2d_gt_for_frame
        or data_selection_settings.require_obb3d_gt_for_frame
    ):
        add_frame_object_info_to_wds(flatten_fsg_dict, wds_dict)
        if data_selection_settings.require_obb2d_gt_for_frame:
            add_frame_object_2d_data_to_wds(flatten_fsg_dict, wds_dict)
        if data_selection_settings.require_obb3d_gt_for_frame:
            add_frame_object_3d_data_to_wds(flatten_fsg_dict, wds_dict)
    if data_selection_settings.require_obb3d_gt_for_frameset:
        add_frameset_object_info_to_wds(flatten_fsg_dict, wds_dict)
        add_frameset_object_3d_data_to_wds(flatten_fsg_dict, wds_dict)
    if data_selection_settings.require_obb3d_gt_for_frameset_group:
        add_frameset_group_object_info_to_wds(flatten_fsg_dict, wds_dict)
        add_frameset_group_object_3d_data_to_wds(flatten_fsg_dict, wds_dict)

    return wds_dict


class AtekWdsWriter:
    """
    An interface to write data to a WebDataset (WDS) directory.
    """

    def __init__(
        self,
        output_path: str,
        prefix_string: str,
        data_selection_settings: DataSelectionSettings,
        max_samples_per_shard: int = 32,
        encryption_handler: Optional[Callable[[bytes], bytes]] = None,
    ):
        """
        Initialize the AtekWdsWriter object.

        Args:
            output_path (str): The path to the output directory where the WDS files will be written.
            prefix_string (str): A prefix string to prepend to the WDS file names.
            data_selection_settings (DataSelectionSettings): Settings for selecting data from the FramesetGroup.
            max_samples_per_shard (int, optional): The maximum number of samples to write to each WDS shard. Defaults to 32.
            encryption_handler (Optional[Callable[[bytes], bytes]], optional): A function to encrypt the WDS data. Defaults to None.
        """
        self.output_path = output_path
        self.prefix_string = prefix_string
        self.settings = data_selection_settings
        self.max_samples_per_shard = max_samples_per_shard
        self.encryption_handler = encryption_handler

        # Initialize sink when writing the first sample
        self.sink = None
        self.next_idx = 0

    def add_sample(self, frameset_group: FramesetGroup):
        """
        Add a sample to the WDS writer.

        Args:
            frameset_group (FramesetGroup): The FramesetGroup object to add to the WDS.
        """
        sample_dict = convert_frameset_group_to_wds_dict(
            self.next_idx, frameset_group, self.prefix_string, self.settings
        )
        if self.encryption_handler is not None:
            sample_dict = encrypt_wds_dict(sample_dict, self.encryption_handler)

        if self.sink is None:
            if not os.path.exists(self.output_path):
                os.makedirs(self.output_path)

            self.sink = wds.ShardWriter(
                f"{self.output_path}/shards-%04d.tar",
                maxcount=self.max_samples_per_shard,
            )

        self.sink.write(sample_dict)
        self.next_idx += 1

    def close(self):
        """
        Close the WDS writer and flush any remaining data to disk.
        """
        if self.sink is not None:
            self.sink.close()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import numpy as np
import pandas as pd
import torch
from atek.utils.obb3 import Obb3
from atek_v1.evaluation.bbox3d_metrics import diagonal_error, iou_giou
from atek_v1.evaluation.math_metrics.distance_metrics import (
    chamfer_distance,
    euclidean_distance,
    hungarian_distance,
)
from atek_v1.evaluation.math_metrics.rotation_metrics import geodesic_angular_error


METRIC_DATA_TYPES = {
    "category_id": int,
    "pred_id": int,
    "gt_id": int,
}


def flatten_to_list(torch_tensor):
    return torch_tensor.flatten().numpy().tolist()


def compute_per_scene_metrics(
    obb3_pred: Optional[Obb3] = None, obb3_gt: Optional[Obb3] = None
) -> pd.DataFrame:
    """
    Compute per-scene pairwise metrics between predicted and GT Obb3's. For example, if there are
    2 predictions and 4 GTs for of the same category, this function will return a dataframe with
    2x4=8 rows. Each row corresponds to one prediction-GT pair. The metrics include intersection
    volume, IoU, and GIoU between each pair of GT and predicted Obb3's.

    Args:
        obb3_pred: Predicted Obb3's
        obb3_gt: Ground truth Obb3's

    Returns:
        metrics_df: DataFrame containing per-scene pairwise metrics
    """

    if obb3_gt is None and obb3_pred is None:
        return pd.DataFrame()
    elif obb3_gt is None:
        pred_num = len(obb3_pred.instance_id)
        metrics = [
            {
                "category_id": flatten_to_list(obb3_pred.category_id),
                "pred_id": flatten_to_list(obb3_pred.instance_id),
                "gt_id": [-1] * pred_num,
                "confidence": flatten_to_list(obb3_pred.score),
                "iou": [np.nan] * pred_num,
                "giou": [np.nan] * pred_num,
                "box_corner_chamfer_distance": [np.nan] * pred_num,
                "box_corner_hungarian_distance": [np.nan] * pred_num,
                "center_euclidean_distance": [np.nan] * pred_num,
                "rotation_geodesic_error": [np.nan] * pred_num,
                "box_corner_diagonal_error": [np.nan] * pred_num,
            }
        ]
    elif obb3_pred is None:
        gt_num = len(obb3_gt.instance_id)
        metrics = [
            {
                "category_id": flatten_to_list(obb3_gt.category_id),
                "pred_id": [-1] * gt_num,
                "gt_id": flatten_to_list(obb3_gt.instance_id),
                "confidence": [np.nan] * gt_num,
                "iou": [np.nan] * gt_num,
                "giou": [np.nan] * gt_num,
                "box_corner_chamfer_distance": [np.nan] * gt_num,
                "box_corner_hungarian_distance": [np.nan] * gt_num,
                "center_euclidean_distance": [np.nan] * gt_num,
                "rotation_geodesic_error": [np.nan] * gt_num,
                "box_corner_diagonal_error": [np.nan] * gt_num,
            }
        ]
    else:
        category_ids = torch.cat((obb3_pred.category_id, obb3_gt.category_id)).unique()
        metrics = []

        for cat_id in category_ids:
            pred_idx = torch.nonzero(obb3_pred.category_id == cat_id).flatten()
            gt_idx = torch.nonzero(obb3_gt.category_id == cat_id).flatten()
            scores = obb3_pred.score[pred_idx]

            if cat_id not in obb3_pred.category_id:
                # False negative: set all metrics to NaN
                gt_num = len(gt_idx)
                curr_metrics = {
                    "category_id": [cat_id] * gt_num,
                    "pred_id": [-1] * gt_num,
                    "gt_id": flatten_to_list(obb3_gt.instance_id[gt_idx]),
                    "confidence": [np.nan] * gt_num,
                    "iou": [np.nan] * gt_num,
                    "giou": [np.nan] * gt_num,
                    "box_corner_chamfer_distance": [np.nan] * gt_num,
                    "box_corner_hungarian_distance": [np.nan] * gt_num,
                    "center_euclidean_distance": [np.nan] * gt_num,
                    "rotation_geodesic_error": [np.nan] * gt_num,
                    "box_corner_diagonal_error": [np.nan] * gt_num,
                }
            elif cat_id not in obb3_gt.category_id:
                # False positive: set all metrics to NaN
                pred_num = len(pred_idx)
                curr_metrics = {
                    "category_id": [cat_id] * pred_num,
                    "pred_id": flatten_to_list(obb3_pred.instance_id[pred_idx]),
                    "gt_id": [-1] * pred_num,
                    "confidence": flatten_to_list(obb3_pred.score[pred_idx]),
                    "iou": [np.nan] * pred_num,
                    "giou": [np.nan] * pred_num,
                    "box_corner_chamfer_distance": [np.nan] * pred_num,
                    "box_corner_hungarian_distance": [np.nan] * pred_num,
                    "center_euclidean_distance": [np.nan] * pred_num,
                    "rotation_geodesic_error": [np.nan] * pred_num,
                    "box_corner_diagonal_error": [np.nan] * pred_num,
                }
            else:
                # compute metrics
                _, iou, giou = iou_giou(
                    obb3_pred.size[pred_idx],
                    obb3_pred.T_ref_obj[pred_idx],
                    obb3_gt.size[gt_idx],
                    obb3_gt.T_ref_obj[gt_idx],
                )
                chamfer_dist = chamfer_distance(
                    obb3_pred.bb3_in_ref_frame[pred_idx],
                    obb3_gt.bb3_in_ref_frame[gt_idx],
                )
                hungarian_dist = hungarian_distance(
                    obb3_pred.bb3_in_ref_frame[pred_idx],
                    obb3_gt.bb3_in_ref_frame[gt_idx],
                )
                euclidean_dist = euclidean_distance(
                    obb3_pred.t_ref_obj[pred_idx],
                    obb3_gt.t_ref_obj[gt_idx],
                )
                geodesic_ang_err = geodesic_angular_error(
                    obb3_pred.R_ref_obj[pred_idx],
                    obb3_gt.R_ref_obj[gt_idx],
                )
                diag_err = diagonal_error(
                    obb3_pred.size[pred_idx],
                    obb3_gt.size[gt_idx],
                )

                pred_id_grid, gt_id_grid = torch.meshgrid(
                    obb3_pred.instance_id[pred_idx], obb3_gt.instance_id[gt_idx]
                )
                confidence_grid, _ = torch.meshgrid(
                    scores, torch.Tensor([1] * len(gt_idx))
                )

                curr_metrics = {
                    "category_id": [cat_id] * len(pred_id_grid.flatten()),
                    "pred_id": flatten_to_list(pred_id_grid),
                    "gt_id": flatten_to_list(gt_id_grid),
                    "confidence": flatten_to_list(confidence_grid),
                    "iou": flatten_to_list(iou),
                    "giou": flatten_to_list(giou),
                    "box_corner_chamfer_distance": flatten_to_list(chamfer_dist),
                    "box_corner_hungarian_distance": flatten_to_list(hungarian_dist),
                    "center_euclidean_distance": flatten_to_list(euclidean_dist),
                    "rotation_geodesic_error": flatten_to_list(geodesic_ang_err),
                    "box_corner_diagonal_error": flatten_to_list(diag_err),
                }
            metrics.append(curr_metrics)

    metrics_df = pd.concat(
        [pd.DataFrame(item).astype(METRIC_DATA_TYPES) for item in metrics],
        ignore_index=True,
    )

    return metrics_df



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List

import numpy as np
import rerun as rr
from projectaria_tools.core.sophus import SE3
from projectaria_tools.utils.rerun import ToTransform3D


TRAJECTORY_COLOR = [30, 100, 30]
GT_COLOR = [30, 200, 30]
PRED_COLOR = [200, 30, 30]


class AtekInferViewer:
    """
    Viewer for the ATEK model inference pipeline, which visualizes model predictions for each
    frame, including RGB image, 3D and 2D bounding boxes. GT 3D and 2D bounding boxes are also
    visualized if available.

    Args:
        web_port (int): port number for the web server
        ws_port (int): port number for the websocket server
        camera_name (str): name of the camera to be visualized
    """

    def __init__(self, web_port: int, ws_port: int, camera_name: str):
        self.camera_name = camera_name
        rr.init("ATEK CubeRCNN Inference Viewer", spawn=True)
        rr.serve(web_port=web_port, ws_port=ws_port)

    def _log_camera_and_image(self, input):
        # image transform: shape CHW -> HWC, color BGR-> RGB
        image = input["image"].detach().cpu().numpy().transpose(1, 2, 0)[:, :, ::-1]

        # log camera
        camera_param = input["K"]
        rr.log(
            f"world/device/{self.camera_name}",
            rr.Pinhole(
                resolution=[
                    image.shape[1],  # width
                    image.shape[0],  # height
                ],
                focal_length=float(camera_param[0][0]),
            ),
            timeless=True,
        )

        # log camera pose
        T_world_cam = SE3.from_matrix3x4(input["T_world_camera"])
        rr.log(
            f"world/device/{self.camera_name}",
            ToTransform3D(T_world_cam, False),
        )

        # log image
        rr.log(
            f"world/device/{self.camera_name}/image",
            rr.Image(image),
        )

    def _log_gt_3d_2d_bbox(self, input):
        # log GT 3D bounding boxes
        bb3ds_sizes_gt = input["object_dimensions"]
        Ts_world_object = [SE3.from_matrix3x4(T) for T in input["Ts_world_object"]]
        bb3ds_centers_gt = [T.translation()[0] for T in Ts_world_object]
        wxyz = [T.rotation().to_quat()[0] for T in Ts_world_object]
        bb3ds_quats_xyzw_gt = [[q[3], q[0], q[1], q[2]] for q in wxyz]
        bb2ds_XYXY = [
            [bb2d[0], bb2d[2], bb2d[1], bb2d[3]] for bb2d in input["bb2ds_x0x1y0y1"]
        ]
        labels_gt = input["category"]
        rr.log(
            f"world/device/bb3d_gt/{self.camera_name}",
            rr.Boxes3D(
                sizes=bb3ds_sizes_gt,
                centers=bb3ds_centers_gt,
                rotations=bb3ds_quats_xyzw_gt,
                radii=0.01,
                colors=GT_COLOR,
                labels=labels_gt,
            ),
        )

        # log GT 2D bounding boxes
        rr.log(
            f"world/device/{self.camera_name}/bb2d_gt",
            rr.Boxes2D(
                array=bb2ds_XYXY,
                array_format=rr.Box2DFormat.XYXY,
                radii=1,
                colors=GT_COLOR,
                labels=labels_gt,
            ),
        )

    def _log_pred_3d_2d_bbox(self, input, prediction):
        T_world_cam = SE3.from_matrix3x4(input["T_world_camera"])
        bb2ds_XYXY_infer = []
        labels_infer = []
        bb3ds_centers_infer = []
        bb3ds_quats_xyzw_infer = []
        bb3ds_sizes_infer = []

        for pred in prediction:
            T_cam_obj_mat = np.zeros([3, 4])
            T_cam_obj_mat[0:3, 0:3] = np.array(pred["R_cam_obj"])
            T_cam_obj_mat[:, 3] = pred["t_cam_obj"]
            T_cam_obj = SE3.from_matrix3x4(T_cam_obj_mat)
            T_world_obj = T_world_cam @ T_cam_obj

            bb2ds_XYXY_infer.append(pred["bbox_2D"])
            labels_infer.append(pred["category"])
            bb3ds_centers_infer.append(T_world_obj.translation()[0])
            wxyz = T_world_obj.rotation().to_quat()[0]
            bb3ds_quats_xyzw_infer.append([wxyz[3], wxyz[0], wxyz[1], wxyz[2]])
            bb3ds_sizes_infer.append(np.array(pred["dimensions"]))

        # log 3D bounding boxes
        rr.log(
            f"world/device/bb3d_infer/{self.camera_name}",
            rr.Boxes3D(
                sizes=bb3ds_sizes_infer,
                centers=bb3ds_centers_infer,
                rotations=bb3ds_quats_xyzw_infer,
                radii=0.01,
                colors=PRED_COLOR,
                labels=labels_infer,
            ),
        )

        # log 2D bounding boxes
        rr.log(
            f"world/device/{self.camera_name}/bb2d_infer",
            rr.Boxes2D(
                array=bb2ds_XYXY_infer,
                array_format=rr.Box2DFormat.XYXY,
                radii=1,
                colors=PRED_COLOR,
                labels=labels_infer,
            ),
        )

    def __call__(self, model_input: List[Dict], model_prediction: List[List[Dict]]):
        assert len(model_input) == len(model_prediction)

        for input, prediction in zip(model_input, model_prediction):
            rr.set_time_nanos("frame_time_ns", input["timestamp_ns"])

            # log pinhole camera and image
            self._log_camera_and_image(input)

            # log ground truth 3D and 2D bounding boxes
            if "Ts_world_object" in input:
                self._log_gt_3d_2d_bbox(input)

            # log predicted 3D and 2D bounding boxes
            if len(prediction) > 0:
                self._log_pred_3d_2d_bbox(input, prediction)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import logging
import math
from typing import List

import numpy as np
import pandas as pd
from atek_v1.data_preprocess.data_schema import Frameset
from atek_v1.data_preprocess.data_utils import (
    check_all_same_member,
    get_rate_stats,
    unify_object_target,
)
from atek_v1.data_preprocess.frame_data_processor import FrameDataProcessor
from atek_v1.data_preprocess.mps_data_processor import MpsDataProcessor

# from atek_v1.utils import mesh_boolean_utils

from projectaria_tools.core.sophus import SE3
from projectaria_tools.core.stream_id import StreamId

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def get_stream_timestamp_key(stream_name: str) -> str:
    return f"{stream_name}_timestamp_ns"


def get_stream_index_key(stream_name: str) -> str:
    return f"{stream_name}_index"


def get_frameset_timestamp(frames_timestamp_ns: List[int], timestamp_type: str):
    """
    Return the timestamp of the frameset computed from the frames timestamp
    based on the timestamp type.
    """
    if timestamp_type == "Average":
        return round(np.mean(frames_timestamp_ns))
    elif timestamp_type == "Earliest":
        return np.min(frames_timestamp_ns)
    else:
        raise ValueError(f"Invalid timestamp type {timestamp_type}")


class FramesetAligner:
    def __init__(
        self,
        target_hz: float,
        frame_data_processors: List[FrameDataProcessor],
        mps_data_processor: MpsDataProcessor = None,
        origin_selection: str = "stream_id#214-1",
        timestamp_type: str = "Average",
        require_objects: bool = False,
    ):
        self.target_hz = target_hz
        self.frame_data_processors = frame_data_processors

        self.mps_data_processor = mps_data_processor

        self.origin_selection = origin_selection
        self.timestamp_type = timestamp_type
        self.require_objects = require_objects

        self.T_device_frameset = self.get_T_device_frameset()

        # Use a data frame to hold various data for fast access and alignment.
        self.frameset_df = self.align_images()
        self.update_df_with_timestamp_info()
        self.update_df_with_mps_info()
        self.cleanup_df()

        self.frameset_fov = None

    def get_T_device_frameset(self):
        """
        Get the device to frameset transform matrix.
        """
        if self.origin_selection.startswith("stream_id"):
            origin_stream_id = StreamId(self.origin_selection.split("#")[-1])
            for frame_data_processor in self.frame_data_processors:
                if frame_data_processor.stream_id == origin_stream_id:
                    T_device_frameset = frame_data_processor.get_T_device_camera()
        elif self.origin_selection == "device":
            T_device_frameset = SE3()
        else:
            raise ValueError(
                f"Not supported origin selection {self.origin_selection}. "
                "The valid options are 'device' or 'stream_id#<stream_id>'"
            )
        return T_device_frameset

    def get_timestamps_ns(self):
        return self.frameset_df["frameset_timestamp_ns"].values

    def update_df_with_timestamp_info(self):
        """
        Update the frameset dataframe with timestamp information.
        """
        stream_timestamp_ns_keys = [
            get_stream_timestamp_key(str(frame_data_processor.stream_id))
            for frame_data_processor in self.frame_data_processors
        ]
        srteam_timestamp_ns_data = self.frameset_df[stream_timestamp_ns_keys].values
        if self.timestamp_type == "Average":
            frameset_timestamp_ns = np.round(
                np.mean(srteam_timestamp_ns_data, axis=-1)
            ).astype(int)
        elif self.timestamp_type == "Earliest":
            frameset_timestamp_ns = np.min(srteam_timestamp_ns_data, axis=-1).astype(
                int
            )
        else:
            raise ValueError(f"Invalid timestamp type {self.timestamp_type}")

        self.frameset_df["frameset_timestamp_ns"] = frameset_timestamp_ns

    def update_df_with_mps_info(self):
        if self.mps_data_processor is not None:
            # Add trajectory info to DataFrame
            T_world_device_dataframe = self.mps_data_processor.get_nearest_poses(
                self.get_timestamps_ns()
            )
            self.frameset_df = pd.concat(
                [self.frameset_df, T_world_device_dataframe], axis=1
            )

            # add semi-dense point cloud info to DataFrame
            semidense_dataframe = self.mps_data_processor.get_nearest_semidense_points(
                self.get_timestamps_ns(), tolerance_ns=150_000_000
            )
            self.frameset_df = pd.concat(
                [self.frameset_df, semidense_dataframe], axis=1
            )

    def cleanup_df(self):
        valid_df = self.frameset_df.dropna()
        invalid_count = len(self.frameset_df) - len(valid_df)
        invalid_percent = (invalid_count / len(self.frameset_df)) * 100
        logger.info(
            f"Dropped {invalid_percent} percent ({invalid_count}/{len(self.frameset_df)}) framesets without required GT information."
        )
        self.frameset_df = valid_df
        self.frameset_df.reset_index(drop=True, inplace=True)

    def align_images(self, tolerance_ns: int = 150000):
        """
        Align the images data of different streams to the first image data stream.
        Only align the nearest images within the tolerance_ns, 150 us is a default
        empirical good number for alignment. Particular consider the default offset
        between rgb image and slam images are 100us.
        """
        # Align image timestamps
        aligned_df = None
        subsample_factor_after_alignment = None
        key_for_align = "timestamp_ns_for_align"
        for frame_data_processor in self.frame_data_processors:
            tss_ns = frame_data_processor.get_timestamps_ns()
            stream_name = str(frame_data_processor.stream_id)

            subsample_factor_after_alignment = (
                min(
                    subsample_factor_after_alignment,
                    frame_data_processor.decide_image_subsample_factor(self.target_hz),
                )
                if subsample_factor_after_alignment is not None
                else frame_data_processor.decide_image_subsample_factor(self.target_hz)
            )

            df = pd.DataFrame(
                {
                    key_for_align: tss_ns,
                    get_stream_timestamp_key(stream_name): tss_ns,
                    get_stream_index_key(stream_name): range(len(tss_ns)),
                }
            )
            if aligned_df is None:
                aligned_df = df
            else:
                aligned_df = (
                    pd.merge_asof(
                        aligned_df.sort_values(key_for_align),
                        df.sort_values(key_for_align),
                        on=key_for_align,
                        tolerance=tolerance_ns,
                        direction="nearest",
                    )
                    .dropna()
                    .astype(int)
                )

        rate_stats = get_rate_stats(aligned_df[key_for_align])
        logger.info(f"Rate status after alignment: {rate_stats}")
        assert math.isclose(
            rate_stats["rate_hz"] / subsample_factor_after_alignment,
            self.target_hz,
            rel_tol=0.01 * self.target_hz,
        )  # allow 1% error

        subsampled_aligned_df = aligned_df.iloc[::subsample_factor_after_alignment]
        rate_stats = get_rate_stats(subsampled_aligned_df[key_for_align])
        logger.info(f"Rate status after subsample: {rate_stats}")

        subsampled_aligned_df.reset_index(inplace=True, drop=True)
        return subsampled_aligned_df

    def update_frameset_fov_mesh(
        self,
        far_clipping_distance: float = 4.0,
        circle_segments: int = 16,
        cap_segments: int = 4,
    ):
        """
        Update and cache a watertight mesh in frameset coordinate
        to represent the field of view of the frameset.
        This mesh could be used for rendering or fov overlapping check.
        """
        self.frameset_fov = None
        for frame_date_processor in self.frame_data_processors:
            frame_date_processor.update_camera_fov_mesh(
                far_clipping_distance, circle_segments, cap_segments
            )
            camera_fov = frame_date_processor.get_camera_fov_mesh()
            T_device_camera = frame_date_processor.get_T_device_camera()
            T_frameset_camera = self.T_device_frameset.inverse() @ T_device_camera
            camera_fov_in_frameset = camera_fov.apply_transform(
                T_frameset_camera.to_matrix()
            )
            if self.frameset_fov is None:
                self.frameset_fov = camera_fov_in_frameset
            else:
                # TODO: temporarily disabled to pass CI. The preprocessing part of atek_v1 should NOT be used!
                """
                self.frameset_fov = mesh_boolean_utils.union_meshes(
                    self.frameset_fov, camera_fov_in_frameset
                )
                """
                pass

    def get_frameset_fov_mesh(self):
        if self.frameset_fov is None:
            # Create a default mesh for the camera fov.
            self.update_frameset_fov_mesh()

        return copy.deepcopy(self.frameset_fov)

    def aligned_frameset_number(self):
        return len(self.frameset_df)

    def check_index_valid(self, index: int):
        assert (
            0 <= index < self.aligned_frameset_number()
        ), f"Index{index} out of bound [0, {self.aligned_frameset_number})."

    def get_frameset_timestamp_by_index(self, index: int) -> int:
        """
        Function to query the frameset timestamp by index.
        """
        self.check_index_valid(index)

        return int(self.frameset_df.loc[index, "frameset_timestamp_ns"])

    def get_T_world_frameset_by_index(self, index: int) -> SE3:
        """
        Function to query the T_world_frameset by index.
        """
        self.check_index_valid(index)
        assert self.mps_data_processor is not None

        T_world_device = SE3.from_quat_and_translation(
            self.frameset_df.loc[index, "qw_world_device"],
            self.frameset_df.loc[
                index, ["qx_world_device", "qy_world_device", "qz_world_device"]
            ],
            self.frameset_df.loc[
                index, ["tx_world_device", "ty_world_device", "tz_world_device"]
            ],
        )
        T_world_frameset = T_world_device @ self.T_device_frameset

        return T_world_frameset

    def get_Ts_frameset_camera_by_index(self, index: int) -> List[SE3]:
        """
        Function to query the Ts_frameset_camera by index.
        """
        self.check_index_valid(index)

        T_frameset_device = self.T_device_frameset.inverse()
        Ts_frameset_camera = [
            (
                T_frameset_device @ frame_data_processor.get_T_device_camera()
            ).to_matrix3x4()
            for frame_data_processor in self.frame_data_processors
        ]

        return Ts_frameset_camera

    def get_frameset_by_index(self, index: int) -> Frameset:
        """
        Get the data by index from all the data processors.
        """
        self.check_index_valid(index)

        frameset = Frameset()
        frameset.frames = []
        for frame_data_processor in self.frame_data_processors:
            stream_name = str(frame_data_processor.stream_id)
            target_frame_index = int(
                self.frameset_df.loc[index, get_stream_index_key(stream_name)]
            )
            frame = frame_data_processor.get_frame_by_index(target_frame_index)
            frameset.frames.append(frame)

        # Fill in the frameset and some sanity checks
        assert len(frameset.frames) > 0, "No valid frames found for frameset."
        assert check_all_same_member(
            frameset.frames, "data_source"
        ), "All frames must be from the same data sources."
        frameset.data_source = frameset.frames[0].data_source

        assert check_all_same_member(
            frameset.frames, "sequence_name"
        ), "All frames must be from the same sequence."
        frameset.sequence_name = frameset.frames[0].sequence_name

        frameset.timestamp_type = self.timestamp_type
        frameset.timestamp_ns = self.get_frameset_timestamp_by_index(index)

        frameset.origin_selection = self.origin_selection
        frameset.Ts_frameset_camera = self.get_Ts_frameset_camera_by_index(index)

        # Generate trajectory information if available.
        if "tx_world_device" in self.frameset_df.columns:
            T_world_frameset = self.get_T_world_frameset_by_index(index)
            frameset.T_world_frameset = T_world_frameset.to_matrix3x4()
            frameset.gravity_in_world = self.frameset_df.loc[
                index, ["gravity_x_world", "gravity_y_world", "gravity_z_world"]
            ].values.reshape((3, 1))
            frameset.gravity_in_frameset = (
                T_world_frameset.rotation().inverse() @ frameset.gravity_in_world
            )

            if self.require_objects:
                unify_object_target(frameset)
                T_frameset_world = T_world_frameset.inverse()
                frameset.Ts_frameset_object = [
                    (
                        T_frameset_world @ SE3.from_matrix3x4(T_world_object)
                    ).to_matrix3x4()
                    for T_world_object in frameset.Ts_world_object
                ]

        # Generate semidense point information if available.
        if "points_world" in self.frameset_df.columns:
            frameset.points_world = self.frameset_df.iloc[index]["points_world"]
            frameset.points_dist_std = self.frameset_df.iloc[index]["points_dist_std"]

        return frameset



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import logging
import math
from typing import Optional

import numpy as np
import pandas as pd

from atek_v1.data_preprocess.base_gt_data_processor import BaseGtDataProcessor
from atek_v1.data_preprocess.data_schema import Frame
from atek_v1.data_preprocess.mps_data_processor import MpsDataProcessor
from atek_v1.utils.camera_utils import get_camera_fov_spherical_cone

from projectaria_tools.core import calibration, data_provider
from projectaria_tools.core.sensor_data import TimeDomain
from projectaria_tools.core.sophus import SE3
from projectaria_tools.core.stream_id import StreamId

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class FrameDataProcessor:
    def __init__(
        self,
        video_vrs: str,
        stream_id: StreamId,
        data_source: str = "unknown",
        time_domain: TimeDomain = TimeDomain.DEVICE_TIME,
        target_linear_camera_params: Optional[np.ndarray] = None,
        rotate_image_cw90deg: bool = True,
        mps_data_processor: Optional[MpsDataProcessor] = None,
        gt_data_processor: Optional[BaseGtDataProcessor] = None,
    ):
        """
        target_framerate: Downsample the frame rate to target_framerate. None means original rate.
        timecode_mapping: DataFrame : timecode mapping from vrs timestamps to unified timestamps
        target_linear_camera_params: Optional[np.adrray] [image_height, image_width, Optional[focal_length_xy]]
            We will use the fxfy inferred from the original camera calibration if it's not provided in target params.
        """

        self.data_source = data_source
        self.time_domain = time_domain
        self.stream_id = stream_id
        self.video_vrs = video_vrs
        self.data_provider = self.setup_data_provider()
        self.mps_data_processor = mps_data_processor

        assert (
            gt_data_processor is None or self.stream_id == gt_data_processor.stream_id
        )
        self.gt_data_processor = gt_data_processor

        # Use a data frame to hold various data for fast access and alignment.
        self.frame_df = pd.DataFrame()
        self.frame_df["timestamp_ns"] = np.array(
            self.data_provider.get_timestamps_ns(self.stream_id, self.time_domain)
        )
        self.frame_df["image_vrs_index"] = range(len(self.frame_df["timestamp_ns"]))
        self.rate_hz = self.data_provider.get_nominalRateHz(self.stream_id)

        stream_sensor_calib = self.data_provider.get_sensor_calibration(self.stream_id)
        assert (
            stream_sensor_calib is not None
        ), f"Can not get the sensor calibration for {self.stream_id} in vrs {video_vrs}"
        self.vrs_camera_calib = stream_sensor_calib.camera_calibration()
        self.final_camera_calib = self.vrs_camera_calib

        self.rotate_image_cw90deg = rotate_image_cw90deg
        if target_linear_camera_params is not None:
            # Use the fxfy inferred from the original camera calibration if it's not provided in target params.
            if len(target_linear_camera_params) > 2:
                fxfy = target_linear_camera_params[2]
            else:
                resize_ratio_W = (
                    target_linear_camera_params[0]
                    / self.vrs_camera_calib.get_image_size()[0]
                )
                resize_ratio_H = (
                    target_linear_camera_params[1]
                    / self.vrs_camera_calib.get_image_size()[1]
                )
                assert math.isclose(
                    resize_ratio_W, resize_ratio_H, rel_tol=1e-4
                ), "Image is resized by different ratio along W and H, can not infer the fxfy for target linear camera"
                fxfy = self.vrs_camera_calib.get_focal_lengths()[0] * resize_ratio_W
                logger.info(f"Inferred the fxfy of target linear camera as {fxfy}.")

            self.target_camera_calib = calibration.get_linear_camera_calibration(
                target_linear_camera_params[0],
                target_linear_camera_params[1],
                fxfy,
                f"target_linear_{self.vrs_camera_calib.get_label()}",
                self.vrs_camera_calib.get_transform_device_camera(),
            )
            self.final_camera_calib = self.target_camera_calib
        else:
            self.target_camera_calib = None

        if self.rotate_image_cw90deg:
            camera_model_to_rotate = (
                self.target_camera_calib
                if self.target_camera_calib is not None
                else self.vrs_camera_calib
            )
            assert (
                camera_model_to_rotate.model_name()
                == calibration.CameraModelType.LINEAR
            ), f"Only support Linear camera model rotation today but got {camera_model_to_rotate.model_name().name}."
            self.rotated_camera_calib = calibration.rotate_camera_calib_cw90deg(
                self.target_camera_calib
            )
            self.final_camera_calib = self.rotated_camera_calib

        self.T_device_camera = self.final_camera_calib.get_transform_device_camera()

        self.update_df_with_mps_info()
        self.cleanup_df()

        if self.gt_data_processor is not None:
            self.gt_data_processor.set_undistortion_params(
                self.vrs_camera_calib,
                self.target_camera_calib,
                self.rotate_image_cw90deg,
            )

        self.camera_fov = None

    def __len__(self):
        return len(self.frame_df)

    def cleanup_df(self):
        valid_df = self.frame_df.dropna()
        invalid_count = len(self.frame_df) - len(valid_df)
        invalid_percent = (invalid_count / len(self.frame_df)) * 100
        logger.info(
            f"Dropped {invalid_percent} percent ({invalid_count}/{len(self.frame_df)}) frames without required GT information."
        )
        self.frame_df = valid_df
        self.frame_df.reset_index(drop=True, inplace=True)

    def get_T_device_camera(self) -> SE3:
        return self.T_device_camera

    def get_rate_hz(self):
        return self.rate_hz

    def get_timestamps_ns(self):
        return self.frame_df["timestamp_ns"].values

    def update_df_with_mps_info(self):
        if self.mps_data_processor is not None:
            # add T_world_device pose info
            T_world_device_dataframe = self.mps_data_processor.get_nearest_poses(
                self.get_timestamps_ns()
            )
            self.frame_df = pd.concat([self.frame_df, T_world_device_dataframe], axis=1)

    def decide_image_subsample_factor(self, target_hz: float):
        # Frame rate sanity check.
        subsample_factor = self.rate_hz / target_hz
        assert math.isclose(subsample_factor, round(subsample_factor), rel_tol=1e-4), (
            f"Can not subsample the image stream {self.stream_id} from {self.rate_hz}hz ",
            f"to {target_hz}hz with the integer subsample factor ({subsample_factor})",
        )
        return int(round(subsample_factor))

    def setup_data_provider(self):
        provider = data_provider.create_vrs_data_provider(self.video_vrs)
        assert provider is not None, f"Cannot open {self.video_vrs}"

        options = provider.get_default_deliver_queued_options()

        all_stream_ids = options.get_stream_ids()
        if self.stream_id not in all_stream_ids:
            logger.error(
                f"{self.stream_id} is not available in {self.video_vrs} which has streams: {all_stream_ids}"
            )

        options.deactivate_stream_all()
        assert options.activate_stream(
            self.stream_id
        ), f"Can not activate stream {self.stream_id} in {self.video_vrs}"

        return provider

    def update_camera_fov_mesh(
        self,
        far_clipping_distance: float = 4.0,
        circle_segments: int = 16,
        cap_segments: int = 4,
    ):
        """
        Update and cache a watertight mesh to represent the field of view of the camera.
        This mesh could be used for rendering or camera fov overlapping check.
        """
        self.camera_fov = get_camera_fov_spherical_cone(
            camera_model=(
                self.final_camera_calib
                if self.final_camera_calib is not None
                else self.vrs_camera_calib
            ),
            far_clipping_distance=far_clipping_distance,
            circle_segments=circle_segments,
            cap_segments=cap_segments,
        )

    def get_camera_fov_mesh(
        self,
    ):
        """
        Return a watertight mesh to represent the field of view of the camera.
        This mesh could be used for rendering or camera fov overlapping check.
        """
        if self.camera_fov is None:
            # Create a default mesh for the camera fov.
            self.update_camera_fov_mesh()

        return copy.deepcopy(self.camera_fov)

    def get_frame_by_index(self, index: int) -> Frame:
        frame = Frame()
        # fill in the input image data
        frame.frame_id = int(self.frame_df["image_vrs_index"][index])
        frame.sequence_name = self.video_vrs
        frame.data_source = self.data_source

        frame.stream_id = str(self.stream_id)
        frame.camera_name = self.vrs_camera_calib.get_label()
        frame.timestamp_ns = int(self.frame_df["timestamp_ns"][index])

        image = self.data_provider.get_image_data_by_index(
            self.stream_id, frame.frame_id
        )[0].to_numpy_array()

        frame.camera_model = self.final_camera_calib.model_name().name
        frame.camera_parameters = self.final_camera_calib.projection_params()

        if self.target_camera_calib is not None:
            # Note that we need to undistort the image first then apply rotation.
            image = calibration.distort_by_calibration(
                image, self.target_camera_calib, self.vrs_camera_calib
            )
            if self.rotate_image_cw90deg:
                image = np.rot90(image, k=-1, axes=(0, 1))

        frame.image = image

        # Generate trajectory information if available.
        if "tx_world_device" in self.frame_df.columns:
            T_world_device = SE3.from_quat_and_translation(
                self.frame_df.loc[index, "qw_world_device"],
                self.frame_df.loc[
                    index, ["qx_world_device", "qy_world_device", "qz_world_device"]
                ],
                self.frame_df.loc[
                    index, ["tx_world_device", "ty_world_device", "tz_world_device"]
                ],
            )
            T_world_camera = T_world_device @ self.T_device_camera
            frame.T_world_camera = T_world_camera.to_matrix3x4()

            frame.gravity_in_world = self.frame_df.loc[
                index, ["gravity_x_world", "gravity_y_world", "gravity_z_world"]
            ].values.reshape(3, 1)
            frame.gravity_in_camera = (
                T_world_camera.rotation().inverse() @ frame.gravity_in_world
            )

        if self.gt_data_processor is not None:
            self.gt_data_processor.get_object_gt_at_timestamp_ns(
                frame, frame.timestamp_ns, T_world_camera.inverse()
            )

        return frame



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Union

import numpy as np
import torch
import trimesh
from projectaria_tools.core import calibration

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def linear_cam_matrix_from_intrinsics(
    camera_params_fufvu0v0: Union[torch.Tensor, np.ndarray]
) -> Union[torch.Tensor, np.ndarray]:
    """
    Generates the camera matrix (nx3x3) for n pinhole camera parameters n x (fx, fy, cx, cy)
    """
    assert camera_params_fufvu0v0.ndim == 2
    assert camera_params_fufvu0v0.shape[-1] == 4
    num_cameras = camera_params_fufvu0v0.shape[0]
    # Initialize an array of zeros
    if isinstance(camera_params_fufvu0v0, torch.Tensor):
        camera_matrices = torch.zeros((num_cameras, 3, 3))
    else:
        camera_matrices = np.zeros((num_cameras, 3, 3))

    # Assign fx, fy, cx, cy
    camera_matrices[:, 0, 0] = camera_params_fufvu0v0[:, 0]  # fx
    camera_matrices[:, 1, 1] = camera_params_fufvu0v0[:, 1]  # fy
    camera_matrices[:, 0, 2] = camera_params_fufvu0v0[:, 2]  # cx
    camera_matrices[:, 1, 2] = camera_params_fufvu0v0[:, 3]  # cy
    camera_matrices[:, 2, 2] = 1.0

    return camera_matrices


def get_camera_fov_cone_angle(
    camera_model: calibration.CameraCalibration,
):
    """Compute the angle of the camera FOV cone in rad."""
    W, H = camera_model.get_image_size()
    principal_point = camera_model.get_principal_point()
    center_ray = np.array([0, 0, 1])

    # Define 4 extremes points to compute the max angle of the camera FOV cone.
    extreme_points = np.array(
        [
            [principal_point[0], 0],
            [principal_point[0], W - 1],
            [0, principal_point[1]],
            [H - 1, principal_point[1]],
        ]
    )
    extreme_points_ray = np.array(
        [
            camera_model.unproject_no_checks(extreme_point)
            for extreme_point in extreme_points
        ]
    )

    # Normalize the extreme points rays
    normalized_rays = extreme_points_ray / np.linalg.norm(
        extreme_points_ray, axis=1, keepdims=True
    )

    # Compute the angles using the dot product and arccos, then convert from radians to degrees
    half_cone_angles = np.arccos(np.dot(normalized_rays, center_ray))

    return half_cone_angles.max()


def create_spherical_cone(
    radius: float, angle: float, circle_segments: int = 32, cap_segments: int = 3
):
    """
    Helper function to create a spherical cone with the given radius and angle.
    Note the cone point is always at the 0, 0, 0, and the cone is aligned with
    the +z axis.
    """
    theta = np.linspace(0, 2 * np.pi, circle_segments)
    phi = np.linspace(0, angle, cap_segments)

    theta, phi = np.meshgrid(theta, phi)

    # Adjust the coordinates to match the desired cone point and angle
    x = radius * np.sin(phi) * np.cos(theta)
    y = radius * np.sin(phi) * np.sin(theta)
    z = radius * np.cos(phi)

    # Create vertices from the coordinate arrays
    vertices = np.column_stack([x.ravel(), y.ravel(), z.ravel()])
    vertices = np.row_stack([np.array([0, 0, 0]), vertices])

    mesh = trimesh.convex.convex_hull(vertices)

    return mesh


def get_camera_fov_spherical_cone(
    camera_model: calibration.CameraCalibration,
    far_clipping_distance: float = 4.0,
    circle_segments: int = 32,
    cap_segments: int = 3,
):
    """
    Helper function to create a spherical cone that represents
    the camera FOV cone. The cone is centered on the camera optical axis.
    """
    half_cone_angle = get_camera_fov_cone_angle(camera_model)
    return create_spherical_cone(
        far_clipping_distance, half_cone_angle, circle_segments, cap_segments
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import json
from functools import partial
from typing import Any, Callable, Dict, List, Optional

import numpy as np

import torch
import webdataset as wds

from atek.util.file_io_utils import merge_tensors_into_dict
from atek.util.tensor_utils import unpack_list_of_tensors


def process_wds_sample(sample: Dict):
    """
    Process one webdataset sample.
    key_process_fn is an optional function helps handling key processing efficiently.
    it takes a key string and return a new processed key string or None after processing.
    None return means we will keep that key for processing.
    TODO: maybe add decryption as Callable
    """
    sample_as_dict = {}
    to_be_stacked_images = {}
    for k, v in sample.items():
        if k in ["__key__", "__local_path__", "__url__"]:
            sample_as_dict[k] = v
        else:
            key_wo_extension, extension_name = k.rsplit(".", 1)
            # Need to restack each image back into tensor of shape `[num_frame, C, H, W]`
            if extension_name == "jpeg":
                # Images are named as `${root_name}_{image_index}.jpeg` in WDS.
                # Break the name into root_name and image_index
                image_root_name = k.split("_")[0]
                image_index = int(k.split("_")[1].split(".")[0])

                if image_root_name not in to_be_stacked_images:
                    to_be_stacked_images[image_root_name] = {}

                # WDS-saved image are loaded as [3, H, W], where single-channel images are simply duplicated for 2 more channels.
                # convert to tensor of shape [F, C, H, W]
                if "rgb" not in image_root_name:
                    image = v[0:1, :, :]
                else:
                    image = v

                # append current image to the correct "stack"
                to_be_stacked_images[image_root_name][image_index] = image

            # Tensors
            elif extension_name == "pth":
                tensor_value = v
                if tensor_value.dtype == torch.float64:
                    tensor_value = tensor_value.float()
                sample_as_dict[key_wo_extension] = tensor_value
            # Dictionary
            elif extension_name == "json":
                sample_as_dict[key_wo_extension] = v
            # string values
            elif extension_name == "txt":
                sample_as_dict[key_wo_extension] = v
            else:
                raise ValueError(f"Unsupported file type in wds {k}")

    # restack images to tensor of [num_frames, C, H, W], but need to re-arrange image order first!
    for image_root_name, img_dict in to_be_stacked_images.items():
        # may have missing frames
        sorted_image_indices = sorted(img_dict.keys())
        image_list_with_correct_order = []

        for index in sorted_image_indices:
            image_list_with_correct_order.append(img_dict[index])

        # Convert each frame to the desired shape
        sample_as_dict[image_root_name] = torch.stack(
            image_list_with_correct_order, dim=0
        )

    # unpack semidense points from a stacked tensor back to List of tensors
    for key in ["points_world", "points_inv_dist_std", "points_dist_std"]:
        if f"msdpd#{key}+stacked" in sample_as_dict:
            sample_as_dict[f"msdpd#{key}"] = unpack_list_of_tensors(
                stacked_tensor=sample_as_dict[f"msdpd#{key}+stacked"],
                lengths_of_tensors=sample_as_dict[f"msdpd#points_world_lengths"],
            )
    # For tensors starting with "GtData#...", merge them back into GT dict
    keys_to_pop = []
    for key, value in sample_as_dict.items():
        if key.startswith("gt_data#") and isinstance(value, torch.Tensor):
            # merge into sample_dict["gt_data"]
            tensor_key_to_be_merged = key.replace("gt_data#", "")
            sample_as_dict["gt_data"] = merge_tensors_into_dict(
                sample_as_dict["gt_data"], {tensor_key_to_be_merged: value}
            )
            keys_to_pop.append(key)

    # pop the original gt tensors from sample_dict
    for tensor_key in keys_to_pop:
        sample_as_dict.pop(tensor_key)

    return sample_as_dict


def select_and_remap_dict_keys(
    sample_dict: Dict[str, Any], key_mapping: Dict[str, str]
) -> Dict[str, Any]:
    """
    Data transform function to modify the sample by selecting and remapping keys according to key_mapping
    """
    remapped_sample = {}
    for k, v in sample_dict.items():
        if k in ["__key__", "__url__"]:
            remapped_sample[k] = v
        elif k in key_mapping:
            new_key = key_mapping[k]
            remapped_sample[new_key] = v
    return remapped_sample


def atek_default_collation_fn(samples):
    """
    Take a collection of samples (dictionaries) from ATEK WDS, and collate them into a batch.
    :param samples: list of sample, each sample is a dict.
    """
    if len(samples) == 0:
        return {}

    assert isinstance(samples[0], dict)

    """
    TODO: add this feature back when needed
    # Get the common keys between samples. This is required when we train with
    # multiple datasets with samples having different keys.
    common_keys = set(samples[0].keys())
    for sample in samples[1:]:
        common_keys &= set(sample.keys())
    """

    batched_dict = {}
    for key in samples[0].keys():
        # first insert values of the same key into a list
        values_as_list = []
        for sample in samples:
            values_as_list.append(sample[key])

        # For tensor list, check if they can be stacked. If so, stack them.
        stackable_flag = False
        if isinstance(values_as_list[0], torch.Tensor):
            stackable_flag = True
            first_tensor_shape = values_as_list[0].shape
            first_tensor_dtype = values_as_list[0].dtype
            for tensor_i in values_as_list:
                if (
                    tensor_i.shape != first_tensor_shape
                    or tensor_i.dtype != first_tensor_dtype
                ):
                    print(
                        f"Tensors for {key} have different shapes or dtypes, cannot be stacked, keep as a list"
                    )
                    stackable_flag = False
                    break

        if stackable_flag:
            batched_dict[key] = torch.stack(values_as_list, dim=0)
        else:
            batched_dict[key] = values_as_list

    return batched_dict


def simple_list_collation_fn(batch):
    # Simply collate as a list
    return list(batch)


def load_atek_wds_dataset(
    urls: List[str],
    nodesplitter: Callable = wds.shardlists.single_node_only,
    dict_key_mapping: Optional[Dict[str, str]] = None,
    data_transform_fn: Optional[Callable] = None,
    batch_size: Optional[int] = None,
    collation_fn: Optional[Callable] = atek_default_collation_fn,
    repeat_flag: bool = False,
    shuffle_flag: bool = False,
) -> wds.WebDataset:
    # 1. load WDS samples back as dicts
    wds_dataset = (
        wds.WebDataset(urls, nodesplitter=nodesplitter)
        .decode(wds.imagehandler("torchrgb8"))
        .map(process_wds_sample)
    )

    # 2. random shuffle
    if shuffle_flag:
        wds_dataset = wds_dataset.shuffle(1000)

    # 3. remap dict keys
    if dict_key_mapping is not None:
        wds_dataset = wds_dataset.map(
            partial(select_and_remap_dict_keys, key_mapping=dict_key_mapping)
        )

    # 4. apply data transforms
    if data_transform_fn is not None:
        wds_dataset = wds_dataset.compose(data_transform_fn)

    # 5. batch samples
    if batch_size is not None:
        wds_dataset = wds_dataset.batched(batch_size, collation_fn=collation_fn)

    # 6. repeat dataset
    if repeat_flag:
        wds_dataset = wds_dataset.repeat()

    return wds_dataset


def create_native_atek_dataloader(
    urls: List[str],
    nodesplitter: Callable = wds.shardlists.single_node_only,
    dict_key_mapping: Optional[Dict[str, str]] = None,
    data_transform_fn: Optional[Callable] = None,
    collation_fn: Optional[Callable] = atek_default_collation_fn,
    batch_size: Optional[int] = None,
    repeat_flag: bool = False,
    shuffle_flag: bool = False,
    num_workers: int = 0,
) -> torch.utils.data.DataLoader:

    wds_dataset = load_atek_wds_dataset(
        urls=urls,
        nodesplitter=nodesplitter,
        dict_key_mapping=dict_key_mapping,
        data_transform_fn=data_transform_fn,
        batch_size=batch_size,
        collation_fn=collation_fn,
        repeat_flag=repeat_flag,
        shuffle_flag=shuffle_flag,
    )

    return torch.utils.data.DataLoader(
        wds_dataset, batch_size=None, num_workers=num_workers, pin_memory=True
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from argparse import Namespace

from atek_v1.model.cubercnn import create_cubercnn_config, create_cubercnn_model


def create_inference_model(args: Namespace):
    """
    Create the model for inference pipeline, with the model config.

    Args:
        args (Namespace):

    Returns:
        model_config: dict-based detailed model configurations
        model: model used for the inference pipeline
    """
    if args.model_name == "cubercnn":
        model_config = create_cubercnn_config(args)
        model = create_cubercnn_model(model_config)
    else:
        raise ValueError(f"Unknown model architecture: {args.model_name}")

    return model_config, model



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from functools import partial

from typing import Callable, Dict, List, Optional

import torch
import webdataset as wds


def identity_key_process(key: str):
    """
    An identity key process function to keep the original keys.
    """
    return key


def convert_tensors_to_float32(obj):
    """
    Recursively converts all tensors in a dictionary, nested dictionary, or list
    from float64 to float32.
    """
    if isinstance(obj, dict):
        return {k: convert_tensors_to_float32(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [convert_tensors_to_float32(item) for item in obj]
    elif isinstance(obj, torch.Tensor) and obj.dtype == torch.float64:
        return obj.float()  # Converts float64 tensor to float32
    else:
        return obj


def process_sample(sample: Dict):
    """
    Process one webdataset sample.
    key_process_fn is an optional function helps handling key processing efficiently.
    it takes a key string and return a new processed key string or None after processing.
    None return means we will keep that key for processing.
    """
    converted_sample = {}
    images = {}
    for k, v in sample.items():
        if k in ["__key__", "__url__"]:
            converted_sample[k] = v
        elif k.endswith(".jpeg"):
            image_key = k.split("_")[0]
            if image_key not in images:
                images[image_key] = []
            images[image_key].append(v)
        elif k.endswith(".pth"):
            assert isinstance(v, dict)
            converted_sample.update(v)
        elif k.endswith(".json"):
            assert isinstance(v, dict)
            converted_sample.update(v)
        else:
            raise ValueError(f"Unsupported file type in wds {k}")

    for k, v in images.items():
        converted_sample[k] = torch.stack(v)

    return convert_tensors_to_float32(converted_sample)


def select_sample_keys(sample, select_key_fn: Callable[[str], bool]):
    """
    Data transform function to modify the sample by poping not selected keys.
    """
    keys_to_pop = []
    for k in sample.keys():
        if k in ["__key__", "__url__"]:
            continue
        elif not select_key_fn(k):
            keys_to_pop.append(k)

    for k in keys_to_pop:
        sample.pop(k)
    return sample


def remap_sample_keys(sample, remap_key_fn: Callable[[str], str]):
    """
    Data transform function to remap the sample keys to user convenient keys.
    """
    remapped_sample = {}
    for k, v in sample.items():
        if k in ["__key__", "__url__"]:
            remapped_sample[k] = v
        else:
            new_key = remap_key_fn(k)
            remapped_sample[new_key] = v
    return remapped_sample


def atek_collation_fn(samples):
    """Take a collection of samples (dictionaries) and create a batch.
    :param dict samples: list of sample
    """
    assert isinstance(samples[0], dict)

    # Get the common keys between samples. This is required when we train with
    # multiple datasets with samples having different keys.
    common_keys = set(samples[0].keys())
    for sample in samples[1:]:
        common_keys &= set(sample.keys())

    batch = {}
    for k in common_keys:
        value = []
        for sample in samples:
            value.append(sample[k])
        if isinstance(value[0], torch.Tensor) and ("object" not in k):
            value = torch.stack(value)
        batch[k] = value

    return batch


def create_atek_webdataset(
    urls,
    batch_size: Optional[int] = None,
    collation_fn: Callable = atek_collation_fn,
    nodesplitter: Callable = wds.shardlists.single_node_only,
    select_key_fn: Optional[Callable[[str], bool]] = None,
    remap_key_fn: Optional[Callable[[str], str]] = None,
    data_transform_fn: Optional[Callable] = None,
    repeat: bool = False,
) -> wds.WebDataset:
    """
    A general standard webdataset processing data pipeline pattern:
    urls-> full atek dict-> [select keys] -> [remap keys] -> [more transform] -> [batch collation]
    Note that we keep the collation in the dataset part for better performance.
    A custom collation function may be required to be compatible with the custom data transform added.

    nodesplitter: set the node splitter properly to handle the multiple node/gpu training. Node that
    wds shardlists splitting function may not work as expected because of different distributed training
    settings. The safest way is to do the splitting manually or write the custom splitting function.
    repeat: set to true makes the dataset to be a infinite iterable dataset.
    """

    dataset = (
        wds.WebDataset(urls, nodesplitter=nodesplitter)
        .decode("torchrgb")
        .map(process_sample)
    )

    if select_key_fn is not None:
        dataset = dataset.map(partial(select_sample_keys, select_key_fn=select_key_fn))
    if remap_key_fn is not None:
        dataset = dataset.map(partial(remap_sample_keys, remap_key_fn=remap_key_fn))
    if data_transform_fn is not None:
        dataset = dataset.compose(data_transform_fn)
    if batch_size is not None:
        dataset = dataset.batched(batch_size, collation_fn=collation_fn)
    if repeat:
        dataset = dataset.repeat()

    return dataset


def create_wds_dataloader(
    atek_dataset: wds.WebDataset,
    num_workers=4,
    pin_memory=False,
):
    """
    Note that the batch size is more efficient to set during creating the webdataset.
    """
    return torch.utils.data.DataLoader(
        atek_dataset,
        batch_size=None,
        num_workers=num_workers,
        pin_memory=pin_memory,
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional

import torch
from atek.data_preprocess.atek_data_sample import MpsOnlineCalibData
from omegaconf.omegaconf import DictConfig
from projectaria_tools.core import mps
from projectaria_tools.core.sensor_data import TimeQueryOptions  # @manual


class MpsOnlineCalibProcessor:
    """
    processor to handle the MPS online calibration data
    """

    def __init__(
        self,
        mps_online_calib_data_file_path: str,
        conf: DictConfig,
    ):
        self.conf = conf

        mps_data_paths = mps.MpsDataPaths()
        mps_data_paths.slam.online_calibrations = mps_online_calib_data_file_path
        self.mps_data_provider = mps.MpsDataProvider(mps_data_paths)

    def get_online_calibration_by_timestamps_ns(
        self, timestamps_ns: List[int]
    ) -> Optional[MpsOnlineCalibData]:
        """
        This function is to get online calibration data by a list of timestamps_ns
        """

        projection_params_list = []  # This will be a list of tensors
        t_device_camera_list = []  # This will be a list of tensors
        capture_timestamp_list = []  # This will store capture timestamps
        utc_timestamp_list = []  # This will store UTC timestamps

        for single_timestamp in timestamps_ns:
            online_calib = self.mps_data_provider.get_online_calibration(
                device_timestamp_ns=single_timestamp,
                time_query_options=TimeQueryOptions.CLOSEST,
            )

            if online_calib is None:
                continue

            capture_timestamps_ns = int(
                online_calib.tracking_timestamp.total_seconds() * 1_000_000_000
            )
            if abs(capture_timestamps_ns - single_timestamp) > self.conf.tolerance_ns:
                continue
            capture_timestamp_list.append(capture_timestamps_ns)
            utc_timestamp_list.append(
                int(online_calib.utc_timestamp.total_seconds() * 1_000_000_000)
            )
            projection_params_per_timestamp_list = []
            t_device_camera_per_timestamp_list = []
            camera_calibs = online_calib.camera_calibs

            for camera_calib in camera_calibs:

                projection_params = camera_calib.projection_params()
                projection_params_per_timestamp_list.append(projection_params)

                t_device_camera_per_timestamp_list.append(
                    camera_calib.get_transform_device_camera().to_matrix3x4().tolist()
                )

            projection_params_list.append(
                torch.tensor(projection_params_per_timestamp_list, dtype=torch.float32)
            )
            t_device_camera_list.append(
                torch.tensor(t_device_camera_per_timestamp_list, dtype=torch.float32)
            )

        if len(projection_params_list) == 0:
            return None

        projection_params_tensor = torch.stack(projection_params_list)
        ts_device_camera_tensor = torch.stack(t_device_camera_list)
        capture_timestamp_tensor = torch.tensor(
            capture_timestamp_list, dtype=torch.int64
        )
        utc_timestamp_tensor = torch.tensor(utc_timestamp_list, dtype=torch.int64)

        return MpsOnlineCalibData(
            capture_timestamps_ns=capture_timestamp_tensor,
            utc_timestamps_ns=utc_timestamp_tensor,
            projection_params=projection_params_tensor,
            ts_device_camera=ts_device_camera_tensor,
        )
